import pandas as pd
from numpy import sqrt
from . import Normalizer
from sklearn.base import BaseEstimator


class simple_clip(BaseEstimator):

    def __init__(self, clip_quantile=0.95):
        """
        A simple clip method to remove outliers. It will clip each feature to their clip_quantile.

        Args:
            clip_quantile (float, optional): . Defaults to 0.95.
        """
        self.clip_quantile = clip_quantile

    def fit(self, x: pd.DataFrame, y: pd.Series = None, sample_weight=None):
        """        
    
        Args:
            x (pandas.DataFrame or a 2D array): The data to compute quantile.
            y (pandas.Series or a 1D array): Null place holder, no effect. Defaults to None.

        Returns:
            self: fitted self
        """
        self.upper = x.quantile(self.clip_quantile)
        self.lower = x.quantile(1 - self.clip_quantile)
        self.fitted_ = True
        return self

    def transform(self, x: pd.DataFrame):
        """
        clip x to upper clip_quantile and lower clip_quantile.

        Args:
            x (pandas.DataFrame or a 2D array): The data to be transformed.

        Returns:
            pandas.DataFrame or a 2D array: transformed x
        """
        if not self.fitted_:
            raise "please call fit before calling transform."

        return x.clip(self.lower, self.upper, axis=1)

    def fit_transform(self,
                      x: pd.DataFrame,
                      y: pd.Series = None,
                      sample_weight=None):
        """
        A functional stack of fit and transform.

        Args:
            x (pandas.DataFrame or a 2D array): The data to compute quantile.
            y (pandas.Series or a 1D array): Null place holder, no effect. Defaults to None.

        Returns:
            pandas.DataFrame or a 2D array: transformed x
        """
        self.fit(x, y)
        x = self.transform(x)
        return x


class IsolationForest(BaseEstimator):
    """
    A wrapper for IsolationForest from sklearn. It will remove outliers based on the IsolationForest algorithm.

    Args:

    """

    def __init__(self):
        pass

    def fit(self, x: pd.DataFrame, y: pd.Series = None, sample_weight=None):
        """Fit the IsolationForest model.

        Args:
            x (pandas.DataFrame or a 2D array): The data to fit the model.
            y (pandas.Series or a 1D array): Null place holder, no effect. Defaults to None.

        Returns:
            self: fitted self
        """
        self.dropped = []
        self.fitted_ = False

        from sklearn.ensemble import IsolationForest

        n_features = x.shape[1]
        self.model = IsolationForest(n_estimators=4 * round(sqrt(n_features)),
                                     max_features=round(sqrt(n_features)),
                                     n_jobs=-1,
                                     random_state=143,
                                     bootstrap=True,
                                     max_samples=0.7)
        self.model.fit(x)
        self.fitted_ = True
        return self

    def transform(self, x: pd.DataFrame):
        """
        Filter the data using the fitted IsolationForest model.

        Args:
            x (pandas.DataFrame or a 2D array): The data to transform.

        Returns:
            pandas.DataFrame or a 2D array: filterd x
        """
        if not self.fitted_:
            raise "please call fit before calling transform."
        abnormal = self.model.predict(x)
        self.dropped.append(x.index[abnormal != 1])
        return x.loc[abnormal == 1]

    def fit_transform(self,
                      x: pd.DataFrame,
                      y: pd.Series = None,
                      sample_weight=None):
        """
        A functional stack of fit and transform.

        Args:
            x (pandas.DataFrame or a 2D array): The data to transform.
            y (pandas.Series or a 1D array): Null placeholder, no effect. Defaults to None.

        Returns:
            pandas.DataFrame or a 2D array: Filterd x
        """
        self.fit(x, y)
        x = self.transform(x)
        return x



from abc import ABC, abstractmethod
from typing import Union

import pandas as pd
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
from seaborn import scatterplot, heatmap, pairplot, color_palette

from sklearn.decomposition import PCA
from sklearn.cross_decomposition import PLSRegression
from sklearn.preprocessing import OneHotEncoder
import sklearn.metrics as metrics


class basic_plot(ABC):
    """
    the base class of plots.    
    Using make_figure(x, y) to generate a figure.    

    """

    def __init__(self,
                 prefix: str = "",
                 save_path: str = "./",
                 save_fig: bool = False,
                 show_fig: bool = True):
        """

        Args:
            prefix (str, optional): the describe or title for the plot. the prefix will be added into the title of plot and saving name. Defaults to "".
            save_path (str, optional): the path to export the figure. Defaults to "./".
            save_fig (bool, optional): whether to export the figure or not. Defaults to False.
            show_fig (bool, optional): whether to show the figure or not. Defaults to True.
        """

        self.prefix = prefix
        self.save_path = save_path
        self.save_fig = save_fig
        self.show_fig = show_fig

    def save_name(self):
        """
        the file name to the saving figure.    

        will be "{prefix} {name}.png"
        """

        return "{} {}".format(self.prefix, self.name)

    def reference(self) -> dict[str, str]:
        """
        This function will return reference of this method in python dict.    
        If you want to access it in PineBioML api document, then click on the    >Expand source code     

        Returns:
            dict[str, str]: a dict of reference.
        """
        refer = {
            "matplotlib document": "https://matplotlib.org/",
            "seaborn document": "https://seaborn.pydata.org/"
        }

        return refer

    @abstractmethod
    def draw(self, x: pd.DataFrame, y: pd.Series = None):
        """
        How and what to draw should be implemented in here.    

        Args:
            x (pd.DataFrame): feature
            y (pd.Series, optional): label. Defaults to None.
        """
        pass

    def make_figure(self, x: pd.DataFrame, y: pd.Series = None):
        """
        1. draw(x, y)    
        2. To save or to show the result.

        Args:
            x (pd.DataFrame): features
            y (pd.Series, optional): label. Defaults to None.
        """
        plt.rcParams.update({'font.family': 'Arial',})
        self.draw(x, y)

        if self.save_fig:
            plt.savefig(self.save_path + self.save_name(), bbox_inches='tight')
        if self.show_fig:
            plt.show()
        else:
            plt.clf()


class pca_plot(basic_plot):
    """
    Calling pca_plot().make_figure(x) or pca_plot().make_figure(x, y) to draw a pca plot of x with n_pc components.    

    """

    def __init__(self,
                 n_pc: int = 4,
                 discrete_legend: bool = True,
                 prefix: str = "",
                 save_path: str = "./output/images/",
                 save_fig: bool = True,
                 show_fig: bool = True):
        """

        Args:
            n_pc (int, optional): number of precipal compoment to plot. Defaults to 4.
            discrete_legend (bool, optional): To color the plot based on y in discrete hue or continuous color bar. If y is continuous, then you should set it to False. Defaults to True.
        """
        super().__init__(prefix=prefix,
                         save_path=save_path,
                         save_fig=save_fig,
                         show_fig=show_fig)
        self.n_pc = n_pc
        self.discrete_legend = discrete_legend
        self.name = "PCA"

    def reference(self) -> dict[str, str]:

        refer = super().reference()
        refer[
            self.name +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html"

        return refer

    def draw(self, x: pd.DataFrame, y: pd.Series = None):
        """
        Using x to draw a pca plot. The difference between pca_plot().draw(x) and pca_plot().draw(x, y) is that:    
         - pca_plot().draw(x) will give a normal pca plot.    
         - pca_plot().draw(x, y) will coloring the points on the figure based on y. Set discrete_legend to True if y is continuous, False otherwise.    

        What we do here is:    
            1. standardize x    
            2. fit a PCA(n_pc) on x and storing the result in pd.DataFrame    
            3. Decide how to color the plots by various cases of y and discrete_legend.    

        Args:
            x (pd.DataFrame): feature    
            y (pd.Series, optional): label. Defaults to None.    
        """

        # calculate pca and store in pd.DataFrame
        pcs = PCA(self.n_pc).fit_transform((x - x.mean()) / (x.std() + 1e-4))
        pcs = pd.DataFrame(
            pcs,
            index=x.index,
            columns=["pc_" + str(i + 1) for i in range(self.n_pc)])

        if y is None:
            y_name = None
        else:
            # add y into the DataFrame for coloring the data points
            y_name = "y" if y.name is None else y.name
            pcs[y_name] = y

        if self.discrete_legend:
            # discrete legend
            plot = pairplot(data=pcs, hue=y_name)
        elif not y is None:
            # color bar
            cmap = color_palette('ch:', as_cmap=True)
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=Normalize())
            sm.set_array([])
            plot = pairplot(data=pcs, hue=y_name)
            plot.legend.remove()

            cbar = plt.colorbar(sm, ax=plot.axes)
            cbar.set_label(y_name)
        else:
            # vanilla
            plot = pairplot(data=pcs)

        plot.figure.suptitle("{} {} scatter plot".format(
            self.prefix, self.name),
                             y=1.01)


class pls_plot(basic_plot):
    """
    PLS-DA is a supervied method in dimension decomposition.    
    This function will plot the result of PLS-DA of given data.    

    Using pls_plot().make_figure(x, y) to generate a figure.

    Warning: PLS-DA is a limited tool in multi-class classification and unlinear regression problem.

    """

    def __init__(self,
                 is_classification: bool,
                 discrete_legend: bool = True,
                 prefix: str = "",
                 save_path: str = "./output/images/",
                 save_fig: bool = True,
                 show_fig: bool = True):
        """

        Args:
            is_classification (bool): If (x, y) is a classification task, then set is_classification to True.
            discrete_legend (bool, optional): To color the plot based on y in discrete hue or continuous color bar. If y is continuous, then you should set it to False. Defaults to True.
        """
        super().__init__(prefix=prefix,
                         save_path=save_path,
                         save_fig=save_fig,
                         show_fig=show_fig)
        self.discrete_legend = discrete_legend
        self.is_classification = is_classification
        self.name = "PLS"

    def reference(self) -> dict[str, str]:

        refer = super().reference()
        refer[
            self.name +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.cross_decomposition.PLSRegression.html"

        return refer

    def draw(self, x: pd.DataFrame, y: pd.Series):
        """
        What we do here is:    
            1. If is_classification => one-hot encode the y. 
            2. Do PLS-DA
            3. Decide how to color the plots by various cases of y and discrete_legend.    

        Args:
            x (pd.DataFrame): features 
            y (pd.Series, optional): label.

        Raises:
            TypeError: _description_
        """
        # one hot encoder for classification
        if self.is_classification:
            OneHot_y = OneHotEncoder(sparse_output=False).fit_transform(
                y.to_numpy().reshape(-1, 1))
            # fit pls regression
            OneHot_y = (OneHot_y - OneHot_y.mean(
                axis=0, keepdims=True)) / OneHot_y.std(axis=0, keepdims=True)
            pls = PLSRegression(n_components=2).fit(x, OneHot_y)
        else:
            if y.dtype == "O":
                raise TypeError(
                    "the dtype of y can't be object while is_classification was setting to False, which means it is a regression task and y should be float or int."
                )
            # fit pls regression
            pls = PLSRegression(n_components=2).fit(x,
                                                    (y - y.mean()) / y.std())

        # project x
        plscs = pls.transform(x)
        plscs = pd.DataFrame(
            plscs,
            index=x.index,
            columns=[self.name + " componet 1", self.name + " componet 2"])

        if y is None:
            y_name = None
        else:
            # add y into the DataFrame for coloring the data points
            y_name = "y" if y.name is None else y.name
            plscs[y_name] = y

        if self.discrete_legend:
            # discrete legend
            plot = scatterplot(data=plscs,
                               x=self.name + " componet 1",
                               y=self.name + " componet 2",
                               hue=y_name)
            plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        elif not y is None:
            # color bar
            cmap = color_palette('ch:', as_cmap=True)
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=Normalize())
            plot = scatterplot(data=plscs,
                               x=self.name + " componet 1",
                               y=self.name + " componet 2",
                               hue=y_name,
                               hue_norm=sm.norm,
                               palette=cmap,
                               legend=False)
            cbar = plt.colorbar(sm, ax=plt.gca())
            cbar.set_label(y_name)
        else:
            # vanilla
            plot = scatterplot(data=plscs,
                               x=self.name + " componet 1",
                               y=self.name + " componet 2")

        plot.set_title("{} {} scatter plot".format(self.prefix, self.name))


class umap_plot(basic_plot):
    """
    Umap is a unsupervissed method to reduce the number of dimension of given data. It's based on manifold learning and it has uncertainty.    

    Using umap_plot().make_figure(x) or umap_plot().make_figure(x, y) to generate a Umap plot.    
    
    Warning: Only the clustering tendency is reliable on the graph. 
    """

    def __init__(self,
                 discrete_legend=True,
                 prefix="",
                 save_path="./output/images/",
                 save_fig=True,
                 show_fig=True):
        """

        Args:
            discrete_legend (bool, optional): To color the plot based on y in discrete hue or continuous color bar. If y is continuous, then you should set it to False. Defaults to True.
        """
        super().__init__(prefix=prefix,
                         save_path=save_path,
                         save_fig=save_fig,
                         show_fig=show_fig)
        self.discrete_legend = discrete_legend
        self.name = "UMAP"

    def reference(self) -> dict[str, str]:

        refer = super().reference()
        refer[self.name +
              " document"] = "https://umap-learn.readthedocs.io/en/latest/"
        refer[
            self.name +
            " publication"] = "https://joss.theoj.org/papers/10.21105/joss.00861"

        return refer

    def draw(self, x: pd.DataFrame, y: pd.Series = None):
        """
        Using Umap do transform x into 2D plane. The difference between umap_plot().draw(x) and umap_plot().draw(x, y) is that:    
         - umap_plot().draw(x) will give a normal umap scatter plot.    
         - umap_plot().draw(x, y) will coloring the points on the figure based on y. Set discrete_legend to True if y is continuous, False otherwise.    

        What we do here is:    
            1. standardize x    
            2. fit a UMAP(n_neighbors=np.log2(n_sample) on x and storing the result in pd.DataFrame    
            3. Decide how to color the plots by various cases of y and discrete_legend.    

        Args:
            x (pd.DataFrame): feature    
            y (pd.Series, optional): label. Defaults to None.    
        """
        from umap import UMAP

        # fit a umap for x, you can change n_neighbors to any other feasible value.
        umapcs = UMAP(n_neighbors=round(np.log2(x.shape[0])),
                      n_components=2).fit_transform(
                          (x - x.mean()) / (x.std() + 1e-6))
        umapcs = pd.DataFrame(
            umapcs,
            index=x.index,
            columns=[self.name + " dimension 1", self.name + " dimension 2"])

        if y is None:
            y_name = None
        else:
            # add y into the DataFrame for coloring the data points
            y_name = "y" if y.name is None else y.name
            umapcs[y_name] = y

        if self.discrete_legend:
            # discrete legend
            plot = scatterplot(data=umapcs,
                               x=self.name + " dimension 1",
                               y=self.name + " dimension 2",
                               hue=y_name)
            plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        elif not y is None:
            # color bar
            cmap = color_palette('ch:', as_cmap=True)
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=Normalize())
            plot = scatterplot(data=umapcs,
                               x=self.name + " dimension 1",
                               y=self.name + " dimension 2",
                               hue=y_name,
                               hue_norm=sm.norm,
                               palette=cmap,
                               legend=False)
            cbar = plt.colorbar(sm, ax=plt.gca())
            cbar.set_label(y_name)
        else:
            # vanilla
            plot = scatterplot(data=umapcs,
                               x=self.name + " dimension 1",
                               y=self.name + " dimension 2")

        plot.set_title("{} {} scatter plot".format(self.prefix, self.name))


class corr_heatmap_plot(basic_plot):
    """
    plot correlation coefficients of given data in heatmap.    

    Using corr_heatmap_plot().make_figure(x) or corr_heatmap_plot().make_figure(x, y) to generate a figure.    
    """

    def __init__(self,
                 prefix="",
                 save_path="./output/images/",
                 save_fig=True,
                 show_fig=True):
        super().__init__(prefix=prefix,
                         save_path=save_path,
                         save_fig=save_fig,
                         show_fig=show_fig)
        self.name = "Correlation Heatmap"

    def draw(self, x: pd.DataFrame, y: pd.Series = None):
        """
        what we do here is:    
            1. add y into x as a new column if y is not None.    
            2. compute correlation between x's columns.
            3. drawint the result in heatmap.

        Args:
            x (pd.DataFrame): feature
            y (pd.Series, optional): Label. Defaults to None.
        """

        data = x.copy()
        if y is None:
            y_name = None
        else:
            # add y into the DataFrame for coloring the data points
            y_name = "y" if y.name is None else y.name
            data[y_name] = y

        plot = heatmap(data.corr(), vmin=-1, vmax=1, cmap='RdBu')

        plot.set_title("{} {}".format(self.prefix, self.name))


class confusion_matrix_plot(basic_plot):
    """
    plot confusion matrix of given ground true label and predictions.
    """

    def __init__(self,
                 prefix="",
                 save_path="./output/images/",
                 save_fig=True,
                 show_fig=True):
        """

        Todo:
            value normalize by y_true and crowding problem in multi-class classification.
        """
        super().__init__(prefix=prefix,
                         save_path=save_path,
                         save_fig=save_fig,
                         show_fig=show_fig)
        self.name = "Confusion Matrix"
        self.normalize = None

    def draw(self, y_true: pd.Series, y_pred: pd.Series):
        """
        plot the confusion matrix using y_true and y_pred.

        Args:
            y_true (pd.Series): Ground true
            y_pred (pd.Series): prediction from an estimator.
        """

        plot = metrics.ConfusionMatrixDisplay.from_predictions(
            y_true,
            y_pred,
            normalize=self.normalize,
            xticks_rotation="vertical")
        plot.ax_.set_title("{} {}".format(self.prefix, self.name))


class roc_plot(basic_plot):
    """
    Depends on how the number of class in y_true and, pos_label, this function will plot a roc curve or several curves of given data.    
    """

    def __init__(self,
                 pos_label: Union[str, int, float] = None,
                 prefix="",
                 save_path="./output/images/",
                 save_fig=True,
                 show_fig=True):
        """

        Args:
            pos_label (Union[str, int, float], optional): If not None, the result will be pos_label vs rest (ovr) roc curve. Defaults to None.
        """
        super().__init__(prefix=prefix,
                         save_path=save_path,
                         save_fig=save_fig,
                         show_fig=show_fig)
        self.name = "ROC Curve"
        self.pos_label = pos_label

    def draw(self, y_true: pd.Series, y_pred_prob: pd.DataFrame):
        """
        draw roc curve

        Args:
            y_true (pd.Series): Ground true
            y_pred_prob (pd.DataFrame): The probability that model predicted for each class. y_pred_prob should have shape (n_samples, n_class)
        """

        # ROC curve
        if len(y_true.value_counts()) <= 2:
            if self.pos_label is None:
                self.pos_label = y_true.iloc[0]
            # binary ROC curve
            fpr, tpr, threshold = metrics.roc_curve(
                y_true, y_pred_prob[self.pos_label], pos_label=self.pos_label)
            roc_auc = metrics.auc(fpr, tpr)
            plt.plot(fpr, tpr, 'b', label='AUC = %0.3f' % roc_auc)
            plt.title(self.prefix + 'ROC curve')

        else:
            # one vs rest ROC curve
            for label in y_pred_prob.columns:
                label_prob = y_pred_prob[label]

                fpr, tpr, threshold = metrics.roc_curve(
                    y_true == label, label_prob)
                roc_auc = metrics.auc(fpr, tpr)

                plt.plot(fpr, tpr, label=str(label) + ' (AUC=%0.3f)' % roc_auc)
            plt.title(self.prefix + 'one vs rest ROC curves')

        plt.plot([0, 1], [0, 1], 'r--')
        plt.xlim([0, 1])
        plt.ylim([0, 1])
        plt.ylabel('True Positive Rate')
        plt.xlabel('False Positive Rate')
        plt.legend(loc='lower right')


def data_overview(input_x: pd.DataFrame,
                  y: pd.Series,
                  is_classification: bool = True,
                  discrete_legend=True,
                  n_pc=4,
                  prefix="",
                  save_fig=True,
                  save_path="./output/images/",
                  show_fig=True):
    """
    Make a glance to data. Specifically it will:    
        1. make a pca plot.    
        2. make a pls plot.    
        3. make a Umap plot.   
        4. make a correlation heatmap.     

    Args:
        input_x (pd.DataFrame): the input feature (or the explanatory variables). 
        y (pd.Series): the target variable (or the response variable).
        is_classification (bool, optional): whether the task to fit y is a classification task or a regression task. Defaults to True.
        discrete_legend (bool, optional): whether to use a discrete legend, otherwise a color bar will be used. If is_classification is True, then discrete_legend will forced to be Ture. Defaults to True.
    """

    x = input_x.copy()
    if is_classification:
        discrete_legend = True

    # PCA
    pca_plot(n_pc=n_pc,
             discrete_legend=discrete_legend,
             prefix=prefix,
             save_path=save_path,
             save_fig=save_fig,
             show_fig=show_fig).make_figure(x, y)

    # PLS
    pls_plot(is_classification=is_classification,
             discrete_legend=discrete_legend,
             prefix=prefix,
             save_path=save_path,
             save_fig=save_fig,
             show_fig=show_fig).make_figure(x, y)

    # UMAP
    umap_plot(discrete_legend=discrete_legend,
              prefix=prefix,
              save_path=save_path,
              save_fig=save_fig,
              show_fig=show_fig).make_figure(x, y)

    # Correlation heatmap
    if y.dtype == "O":
        corr_heatmap_plot(prefix=prefix,
                          save_path=save_path,
                          save_fig=save_fig,
                          show_fig=show_fig).make_figure(x)
    else:
        corr_heatmap_plot(prefix=prefix,
                          save_path=save_path,
                          save_fig=save_fig,
                          show_fig=show_fig).make_figure(x, y)


def classification_summary(y_true,
                           y_pred_prob,
                           target_label=None,
                           prefix="",
                           save_path="./output/images/",
                           save_fig=True,
                           show_fig=True):
    """
    Give a classification summary, including:    
        1. recall, precision, f1 and accuracy.    
        2. confusion matrix.    
        3. ROC curve.    

    Args:
        y_true (pandas.Series or a 1D array): The label.
        y_pred_prob (pandas.Series or a 1D array): float in [0, 1]. prediction from model.
        target_label ( member of label, optional): The target class from y_true to compute sensitivity and specificity. Defaults to None.
    Todo:
        1. support to multi-class classification(on going)
        2. the label matching between y_true and y_pred
        3. support to regression
    """
    y_pred = y_pred_prob.idxmax(axis=1)
    #print("\n", prefix)
    #print(metrics.classification_report(y_true, y_pred))

    report = metrics.classification_report(y_true, y_pred, output_dict=True)
    acc = report.pop("accuracy")
    report = pd.DataFrame(report).T
    report.loc["accuracy"] = [" ", " ", acc, y_true.count()]

    if not target_label is None:
        # binary classification
        prf1 = metrics.classification_report(y_true == target_label,
                                             y_pred == target_label,
                                             output_dict=True)

        report.loc["sensitivity"] = [
            " ", " ", prf1["True"]["recall"], prf1["True"]["support"]
        ]
        report.loc["specificity"] = [
            " ", " ", prf1["False"]["recall"], prf1["True"]["support"]
        ]

        #print("sensitivity: {:.3f}".format(sensitivity))
        #print("specificity: {:.3f}".format(specificity))

    # insert an empty row to split
    tmp = report.loc[y_pred_prob.columns.astype(str)]
    tmp.loc["   "] = [" ", " ", " ", " "]
    report = pd.concat([tmp, report.drop(y_pred_prob.columns.astype(str))],
                       axis=0)

    print(report)
    if save_fig:
        report.to_csv(save_path + "{} scores.csv".format(prefix))

    # confusion matrix
    confusion_matrix_plot(prefix=prefix,
                          save_path=save_path,
                          show_fig=show_fig,
                          save_fig=save_fig).make_figure(y_true, y_pred)

    # roc cuve
    roc_plot(pos_label=target_label,
             prefix=prefix,
             save_path=save_path,
             show_fig=show_fig,
             save_fig=save_fig).make_figure(y_true, y_pred_prob)


def regression_summary(y_true: pd.Series,
                       y_pred: pd.Series,
                       x: pd.DataFrame = None,
                       prefix="",
                       save_path="./output/images/",
                       save_fig=True,
                       show_fig=True):
    """
    1. compute rmse, r square and mape if y is all positive.    
    2. feature pca residual plot if x is not None.

    Args:
        y_true (pd.Series): Ground true
        y_pred (pd.Series): The estimates from model.
        x (pd.DataFrame, optional): features. If not None, it will be used to plot a feature pca residual plot. Defaults to None.
    """

    scores = pd.Series(
        {
            "rmse": metrics.root_mean_squared_error(y_true, y_pred),
            "r2": metrics.r2_score(y_true, y_pred),
            "mae": metrics.mean_absolute_error(y_true, y_pred),
        },
        name="performance")
    if (y_true > 0).all():
        scores["mape"] = metrics.mean_absolute_percentage_error(y_true, y_pred)
    scores["support"] = len(y_true)

    print("\n", prefix, " performance:")
    print(scores)
    if save_fig:
        scores.to_csv(save_path + "{} scores.csv".format(prefix))

    residual = y_true - y_pred
    if not x is None:
        pca_plot(n_pc=2,
                 discrete_legend=False,
                 prefix=prefix + " Residual",
                 save_path=save_path,
                 save_fig=save_fig,
                 show_fig=show_fig).make_figure(x, residual)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import t
from . import SelectionPipeline


class Volcano_selection(SelectionPipeline):
    """
    volcano plot.

    """

    def __init__(self,
                 k,
                 z_importance_threshold=1.,
                 strategy="fold",
                 p_threshold=0.05,
                 fc_threshold=2,
                 log_domain=False,
                 absolute=True,
                 target_label=1):
        """

        Args:
            strategy (str, optional): Choosing strategy. One of {"p" or "fold"} Defaults to "fold".
            p_threshold (float, optional): p-value threshold. Only feature has p-value higher than threshold will be considered. Defaults to 0.05.
            fc_threshold (int, optional): fold change threshold. Only feature has fold change higher than threshold will be considered. Defaults to 2.
            log_domain (bool, optional): Whether input data is in log_domain. Defaults to False.
            absolute (bool, optional): If true, then take absolute value on score while strategy == "p". Defaults to True.
            target_label : the target label.
        """
        super().__init__(k=k, z_importance_threshold=z_importance_threshold)
        self.strategy = strategy
        self.fc_threshold = fc_threshold
        self.p_threshold = p_threshold
        self.log_domain = log_domain
        self.absolute = absolute
        self.name = "Volcano Plot_" + self.strategy
        self.missing_value = 0
        self.target_label = target_label

    def Scoring(self, x, y):
        """
        Compute the fold change and p-value on each feature.

        Args:
            x (pandas.DataFrame or a 2D array): The data to extract information.
            y (pandas.Series or a 1D array): The target label for methods. Defaults to None.

        Returns:
           pandas.DataFrame: A dataframe records p-value and fold change.
        """
        positive = y == self.target_label
        negative = np.logical_not(positive)

        x = x.replace(0, np.nan)

        # fold change
        if not self.log_domain:
            log_fold = np.log2(x[positive].mean(axis=0) /
                               x[negative].mean(axis=0))
        else:
            log_fold = x[positive].mean(axis=0) - x[negative].mean(axis=0)

        # Welch t test:
        #     normal assumption
        #     diffirent sample size
        #     diffirent varience
        #     unpaired
        n_positive = x[positive].shape[0]
        n_negative = x[negative].shape[0]
        diff = x[positive].mean(axis=0) - x[negative].mean(axis=0)

        s_positive = ((x[positive] - x[positive].mean(axis=0))**
                      2).sum(axis=0) / (n_positive - 1)
        s_negative = ((x[negative] - x[negative].mean(axis=0))**
                      2).sum(axis=0) / (n_negative - 1)
        st = np.sqrt(s_positive / n_positive + s_negative / n_negative)

        t_statistic = np.abs(diff / st)
        df = (s_positive / n_positive + s_negative / n_negative)**2 / (
            (s_positive / n_positive)**2 / (n_positive - 1) +
            (s_negative / n_negative)**2 / (n_negative - 1))

        # 2 side testing
        #print("t statistic: ", t_statistic.min(), t_statistic.mean(), t_statistic.max())
        #print("degree of freedom: ", df.min(), df.mean(), df.max())

        p_value = t.cdf(x=-t_statistic, df=df) * 2
        log_p = -np.log10(p_value)

        scores = pd.DataFrame(
            {
                "log_p_value": log_p,
                "log_fold_change": log_fold
            },
            index=log_fold.index)
        return scores

    def Select(self, scores):
        """
        Choosing the features which has score higher than threshold in assigned strategy.

        If strategy == "fold": sort in fold change and return p-value

        If strategy == "p": sort in p-value and return fold change

        Args:
            scores (pandas.DataFrame): A dataframe records p-value and fold change.
            k (int): Number of features to select.

        Returns:
            pandas.Series: The score for k selected features in assigned strategy.
        """
        log_fold = scores["log_fold_change"]
        log_p = scores["log_p_value"]
        # choose fold change > 2 and p value < 0.05 in log scale
        significant = np.logical_and(
            np.abs(log_fold) >= np.log2(self.fc_threshold), log_p
            > -np.log10(self.p_threshold))
        self.significant = significant

        # choose top k logged p-value
        if self.strategy == "fold":
            selected = np.abs(log_fold).loc[significant].sort_values().tail(
                self.k)
            selected_score = pd.Series(log_p.loc[selected.index],
                                       index=selected.index,
                                       name=self.name)
        elif self.strategy == "p":
            selected = log_p.loc[significant].sort_values().tail(self.k)
            if self.absolute:
                selected_score = pd.Series(
                    np.abs(log_fold.loc[selected.index]),
                    index=selected.index,
                    name=self.name).sort_values(ascending=False)
            else:
                selected_score = pd.Series(
                    log_fold.loc[selected.index],
                    index=selected.index,
                    name=self.name).sort_values(ascending=False)
        else:
            raise "select_by must be one of {fold} or {p}"

        return selected_score

    def plotting(self,
                 external=False,
                 external_score=None,
                 title="Welch t-test volcano",
                 show=True,
                 saving=False,
                 save_path="./output/"):
        """
        Plotting

        Args:
            external (bool, optional): True to use external score. Defaults to False.
            external_score (_type_, optional): External score to be used. Only activate when external == True. Defaults to None.
            title (str, optional): plot title. Defaults to "Welch t-test volcano".
            show (bool, optional): True to show the plot. Defaults to True.
            saving (bool, optional): True to save the plot. Defaults to False.
            save_path (str, optional): The path to save plot. Only activate when saving == True. Defaults to "./output/images/".
        """
        log_fold = self.scores_["log_fold_change"]
        log_p = self.scores_["log_p_value"]
        # choose fold change > 2 and p value < 0.05 in log scale
        significant = np.logical_and(
            np.abs(log_fold) >= np.log2(self.fc_threshold), log_p
            > -np.log10(self.p_threshold))

        if external:
            selected = pd.Series(False, index=self.scores_.index)
            selected.loc[external_score.index] = True
        else:
            selected = pd.Series(False, index=significant.index)
            selected.loc[self.selected_score_.index] = True

        # silent
        plt.scatter(x=log_fold[~significant],
                    y=log_p[~significant],
                    s=0.5,
                    color='gray')
        # not selected
        plt.scatter(x=log_fold[significant][~selected],
                    y=log_p[significant][~selected],
                    s=2)
        # selected
        if external:
            plt.scatter(x=log_fold[selected], y=log_p[selected], s=2)
        else:
            plt.scatter(x=log_fold[selected], y=log_p[selected], s=2)

        plt.title(title)
        plt.xlabel("log_2 fold")
        plt.axvline(1, linestyle="dotted", color="gray")
        plt.axvline(-1, linestyle="dotted", color="gray")
        plt.axhline(-np.log10(0.05), linestyle="dotted", color="gray")
        plt.ylabel("log_10 p")
        if saving:
            plt.savefig(save_path + title, format="png")
        plt.show()



import warnings
import sklearn.metrics as metrics
from sklearn.base import BaseEstimator
import optuna
from abc import ABC, abstractmethod
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.preprocessing import LabelEncoder
from optuna.samplers import TPESampler
from numpy.random import RandomState, randint
from numpy import zeros
from pandas import Series, DataFrame, concat
from shap import Explainer

from sklearn.exceptions import ConvergenceWarning

optuna.logging.set_verbosity(optuna.logging.ERROR)
ConvergenceWarning('ignore')
"""
class Binary_threshold_wrapper(ClassifierMixin):

    def __init__(self, kernel, threshold):
        self.threshold = threshold
        self.kernel = kernel
        self.classes_ = kernel.classes_

    def fit(self, x, y):
        pass

    def predict(self, x):
        decision = self.kernel.predict_proba(x)[:, 1] > self.threshold
        return decision.astype(int)

    def predict_proba(self, x):
        return self.kernel.predict_proba(x)
"""


class Basic_tuner(ABC, BaseEstimator):
    """
    The base class of tuner. A tuner is a wrapper of optuna + models    

    What the tuners do:    
        1. interface of optuna and models with sklearn api style    
        2. randomity management.    
        3. providing a uniform interface to regression and classiciation(binary and multi-class)    
    To conserve the reproducibility and to reduce the hyper parameter overfitting along the process of hyper parameter tuning,    
    we first using valid_seed to randomly initialize a tape of integers and it will sequentially be used in optuna trials.    
    """

    def __init__(self,
                 n_try: int,
                 n_cv: int,
                 target: str,
                 validate_penalty: bool,
                 TT_coef: float,
                 kernel_seed: int = None,
                 valid_seed: int = None,
                 optuna_seed: int = None):
        """

        Args:
            n_try (int): The number of trials optuna should try.
            n_cv (int): The number of folds to execute cross validation evaluation in iteration of optuna optimization.
            target (str): The target of optuna optimization. Notice that is different from the training loss of model.
            validate_penalty (bool): Deprecated.
            TT_coef (float): The power of penalty to overfitting by Tibshirani & Tibshirani method. Ranges in [0, 1]. Add the difference between training score and cv score to the optimization target.
            kernel_seed (int, optional): Random seed for model. Defaults to None.
            valid_seed (int, optional): Random seed for cross validation. Defaults to None.
            optuna_seed (int, optional): Random seed for optuna's hyperparameter sampling. Defaults to None.

        ToDo:
            1. transfer the initialization of seed tape from __init__ to fit.    
            2. optuna pruner: See the section Acticating Pruners in https://optuna.readthedocs.io/en/stable/tutorial/10_key_features/003_efficient_optimization_algorithms.html
            3. Winner's curse

        """
        self.validate_penalty = validate_penalty
        self.TT_coef = TT_coef
        self.n_cv = n_cv
        self.n_try = n_try
        self.target = target
        self.kernel_seed = kernel_seed
        self.valid_seed = valid_seed
        self.optuna_seed = optuna_seed
        
    
    def _set_up(self):
        if self.validate_penalty == True:
            warnings.warn(
                "validate_penalty will be remove in future. Use argument TT_coef.",
                DeprecationWarning,
                stacklevel=2)
            
        self.y_mapping = LabelEncoder()
        self.fitted_ = None
        self.optuna_early_stop_counter = self.n_cv // 10 + 2
        self.n_sample = 1
        self.n_opt_jobs = 1
        self.default = False
        self.training = True

        # initialize the random seeds
        if self.kernel_seed is None:
            self.kernel_seed = randint(16384)
        else:
            self.kernel_seed = self.kernel_seed

        if self.valid_seed is None:
            self.valid_seed = randint(16384)
        else:
            self.valid_seed = self.valid_seed

        if self.optuna_seed is None:
            self.optuna_seed = randint(16384)
        else:
            self.optuna_seed = self.optuna_seed

        # The random seed tapes for cross validation along the optuna's optimization trials.
        self.valid_seed_tape = RandomState(self.valid_seed).randint(low=0, high=16384, size=self.n_try)
        self.kernel_seed_tape = RandomState(self.kernel_seed).randint(low=0, high=16384, size=self.n_try)

        self.metric = self.get_scorer(self.target)

        # the model optuna tuned
        self.optuna_model = None
        # the model with default params
        self.default_model = None
        # the better one of self.default_model or self.optuna_model
        self.best_model = None
        # the thresholds for binary classification along optuna tuning
        #self.thresholds = zeros(self.n_try) + 0.5
        # the threshold coresponding to the best optuna trial
        #self.best_threshold = 0.5
        # the stopping point of early stopping for Boosting methods.
        self.stop_points = zeros(self.n_try, dtype=int)

        self.explainer = None

    def reference(self) -> dict[str, str]:
        """
        This function will return reference of this method in python dict.    
        If you want to access it in PineBioML api document, then click on the    >Expand source code     

        Returns:
            dict[str, str]: a dict of reference.
        """
        refers = {
            "optuna publication":
            "https://dl.acm.org/doi/10.1145/3292500.3330701",
            "optuna document": "https://optuna.org/",
            "sklearn publication":
            "https://dl.acm.org/doi/10.5555/1953048.2078195"
        }

        return refers

    @abstractmethod
    def name(self) -> str:
        """
        To be determined.

        Returns:
            str: Name of this tuner.
        """
        pass

    @abstractmethod
    def is_regression(self) -> bool:
        """
        Returns:
            bool: True if the task of tuner is a regression task.
        """
        pass

    def using_earlystopping(self) -> bool:
        """
        Returns:
            bool: True if applying earlystopping in optimizing training.
        """
        return False

    def accepting_categorical_features(self) -> bool:
        """
        Returns:
            bool: Ture if kernel receives categorical (discreate) features.
        """
        return False

    @abstractmethod
    def parms_range(self) -> dict:
        """
        model hyper-parameter search range.

        Returns:
            dict: {parameter_name : (parameter_name, parameter_dtype, lower_bound, upper_bound)}
        """
        pass

    def parms_range_sparser(self, trial, search_setting):
        parameter_name, parameter_dtype, lower_bound, upper_bound = search_setting
        param = None

        if parameter_dtype in ["float", "int"]:
            log = lower_bound > 0 and upper_bound / lower_bound > 10
            if parameter_dtype == "float":
                param = trial.suggest_float(parameter_name,
                                            lower_bound,
                                            upper_bound,
                                            log=log)
            elif parameter_dtype == "int":
                param = trial.suggest_int(parameter_name,
                                          lower_bound,
                                          upper_bound,
                                          log=log)
        else:
            if parameter_dtype == "bool":
                param = trial.suggest_categorical(parameter_name,
                                                  [lower_bound, upper_bound])
            elif parameter_dtype == "category":
                param = trial.suggest_categorical(parameter_name, lower_bound)
        if param is None:
            raise ValueError(
                "parameter type not support, receaive parameter_dtype {}, parameter_name {}. Only one of [\"float, int, bool, category] is supported."
                .format(parameter_dtype, parameter_name))
        return param

    @abstractmethod
    def create_model(self, trial, default, training) -> BaseEstimator:
        """
        Create model based on default setting or optuna trial over search range.

        Args:
            trial (optuna.trial.Trial): optuna trial in this call.
            default (bool): set True to use default hyper parameter
            training (bool): set True to when training.
            
        Returns :
            sklearn.base.BaseEstimator: A sklearn style model object.
        """
        pass

    def optimize_fit(self, clr, train_data, sample_weight, valid_data):
        """
        optimize_fit the polymorphism middle layer between model fitting and optuna optimize evaluate.    
        Specifically, optimize_fit is used for XGBoost/lightGBM/Catboost early stopping which requires validation data in .fit .
        However sklearn estimators such as randomforest does not provide such api.    

        Args:
            clr (classifier): classifier to fit.
            train_data (tuple): (train_x, train_y) .
            sample_weight (list-like): training sample weight.
            valid_data (tuple): (valid_x, valid_y) .

        Returns:
            sklearn.base.BaseEstimator: A fitted sklearn style model object.
        """
        train_x, train_y = train_data
        valid_x, valid_y = valid_data

        return clr.fit(train_x, train_y, sample_weight=sample_weight)

    def clr_best_iteration(self, classifier) -> int:
        """
        the polymorphism to call classifier's early stopping rounds/iteration/point.    
        only need to be overide when subclass is implemented with early stopping.

        Args:
            classifier: the classifier object.

        Returns:
            int: the stopping point/rounds/iteration.
        """
        pass

    def evaluate(self, trial, default=None, training=None) -> float:
        """
        To evaluate the score of this trial. you should call create_model instead of creating model manually in this function.
        
        Args:
            trial (optuna.trial.Trial): optuna trial in this call.
            default (bool): To use default hyper parameter. This argument will be passed to creat_model .    
            training (bool): whether it is under optuna optimization or not. False then evalutate will not tune a threshold or applying early stop. 
        Returns :
            float: The score. Decided by optimization target.
        """
        if default is None:
            default = self.default
        if training is None:
            training = self.training

        # create the model using from this trial
        classifier_obj = self.create_model(trial, default, training=True)

        # create cross validation
        if self.is_regression():
            cv = KFold(n_splits=self.n_cv,
                       shuffle=True,
                       random_state=self.valid_seed_tape[trial.number])
        else:
            cv = StratifiedKFold(
                n_splits=self.n_cv,
                shuffle=True,
                random_state=self.valid_seed_tape[trial.number])

        # do cv
        score = zeros(self.n_cv)
        # thresholds not None only if  self.is_binary and not self.is_regression
        cv_thresholds = zeros(self.n_cv) + 0.5
        cv_stoppoint = zeros(self.n_cv)
        for i, (train_ind, test_ind) in enumerate(cv.split(self.x, self.y)):
            # train test split
            x_train = self.x.iloc[train_ind]
            y_train = self.y.iloc[train_ind]
            x_test = self.x.iloc[test_ind]
            y_test = self.y.iloc[test_ind]

            # sample_weight for imbalanced class.
            if self.is_regression():
                sample_weight = None
            else:
                sample_weight = compute_sample_weight(class_weight="balanced",
                                                      y=y_train)

            # fit the model on training fold
            fitted_clr = self.optimize_fit(clr=classifier_obj,
                                           train_data=(x_train, y_train),
                                           sample_weight=sample_weight,
                                           valid_data=(x_test, y_test))

            # record the stopping point of early stopping.
            if self.using_earlystopping() and training:
                cv_stoppoint[i] = self.clr_best_iteration(fitted_clr)

            # tune a threshold via roc for Binary classification
            """
            if self.is_binary and not self.is_regression():
                if training:
                    fpr, tpr, thr = metrics.roc_curve(
                        y_test,
                        fitted_clr.predict_proba(x_test)[:, 1])
                    ### TODO: flexible threshold picker for various metrics.
                    cv_thresholds[i] = thr[abs(tpr - fpr).argmax()]
                else:
                    cv_thresholds[i] = self.thresholds[trial.number]
                # wrap the classifier before calculate scores
                if not default:
                    fitted_clr = Binary_threshold_wrapper(
                        fitted_clr, cv_thresholds[i])
                else:
                    fitted_clr = Binary_threshold_wrapper(fitted_clr, 0.5)
            """

            # evaluate on testing fold
            test_score = self.metric(fitted_clr, x_test, y_test)
            train_score = self.metric(fitted_clr, x_train, y_train)

            score[i] = test_score + self.TT_coef * (test_score - train_score)

        # averaging over cv
        if training:
            #self.thresholds[trial.number] = cv_thresholds.sum() / self.n_cv
            self.stop_points[trial.number] = round(cv_stoppoint.sum() /
                                                   (self.n_cv - 1))
        return score.mean()

    def tune(self, x, y) -> None:
        """
        this function tunes the hyperparameters of a given kernel.
        
        Args:
            x (pandas.DataFrame or 2D-array): feature to extract information from.
            y (pandas.Series or 1D-array): The property we interested in.

        """
        # input data
        self.x = x
        self.y = y

        # the total number of samples
        self.n_sample = x.shape[0]

        # check task
        self.check_task()

        # Make the sampler behave in a deterministic way.
        sampler = TPESampler(seed=self.optuna_seed,
                             multivariate=True,
                             n_startup_trials=self.n_try // 3)
        self.study = optuna.create_study(direction="maximize", sampler=sampler)

        print(
            "optuna seed {self.optuna_seed}  |  validation seed {self.valid_seed}  |  model seed {self.kernel_seed}"
            .format(self=self))
        print("    {} start tuning. it will take a while.".format(self.name()))
        # using optuna tuning hyper parameter
        self.training = True
        self.default = False
        self.study.optimize(self.evaluate,
                            n_trials=self.n_try,
                            show_progress_bar=False,
                            callbacks=[self.early_stop_callback],
                            n_jobs=self.n_opt_jobs)
        self.optuna_model = self.create_model(self.study.best_trial,
                                              default=False,
                                              training=False)

        # using default hyper parameter
        # the best_trial here is only a placeholder. It's not functional.
        self.default_performance = self.evaluate(self.study.best_trial,
                                                 default=True,
                                                 training=False)
        self.default_model = self.create_model(self.study.best_trial,
                                               default=True,
                                               training=False)
        #print("    default performance: {:.3f}  |  best performance: {:.3f}".
        #      format(self.default_performance, self.study.best_trial.value))
        if self.default_performance > self.study.best_trial.value:
            # default better
            print("    default is better.")
            self.best_model = self.default_model
        else:
            # optuna better
            print("    optuna is better, best trial: ",
                  self.study.best_trial.number)
            self.best_model = self.optuna_model
            """
            # threshold tuner for binary classification
            if self.is_binary and not self.is_regression():
                self.best_threshold = self.thresholds[
                    self.study.best_trial.number]
            """

        # release the datas
        self.x = None
        self.y = None

    def fit(self, x, y, retune=True):
        """
        The sklearn.base.BaseEstimator fit api.    
        Set retune to True to retune the model using the given x and y, else using the previously tuned model to fit on given x and y.

        Args:
            x (pandas.DataFrame or 2D-array): feature to extract information from.
            y (pandas.Series or 1D-array): ground true.
            retune (bool): True to retune the model using given x and y, else using the tuned model to fit on given x, y.
        """
        if retune:
            self._set_up()
            
        if hasattr(y, "name"):
            self.label_name_ = y.name
        else:
            self.label_name_ = None

        # label encoding
        if hasattr(y, "index"):
            idx = y.index
        else:
            idx = None
        if not self.is_regression():
            y = Series(self.y_mapping.fit_transform(y),
                       index=idx,
                       name=self.label_name_)

        # tune the model.
        if retune:
            self.tune(x, y)

        # fit the model.
        if self.is_regression():
            sample_weight = None
        else:
            sample_weight = compute_sample_weight(class_weight="balanced", y=y)
        self.best_model.fit(x, y, sample_weight=sample_weight)

        self.fitted_ = True

        return self

    def predict(self, x):
        """
        The sklearn.base.BaseEstimator predict api.

        Args:
            x (pandas.DataFrame or 2D-array): feature to extract information from.

        Returns:
            1D-array: prediction
        """
        # using the model.
        """
        if self.is_binary and not self.is_regression():
            y_pred = (self.best_model.predict_proba(x)[:, 1]
                      > self.best_threshold).astype(int)
        else:
        """
        y_pred = self.best_model.predict(x)

        # label encoding
        if not self.is_regression():
            y_pred = self.y_mapping.inverse_transform(y_pred)
        if hasattr(x, "index"):
            idx = x.index
        else:
            idx = None
            
        y_pred = Series(y_pred, index=idx, name=self.label_name_)

        return y_pred

    def get_scorer(self, scorer_name: str):
        """
        A lazy function to call sklearn scorers.    

        Args:
            scorer_name (str): abbreviation or formal name of sklearn scorers.

        Returns:
            scorer: A callable object that returns score.
        """

        ### polymorphism
        scorer_name = scorer_name.lower().replace("-", "_").replace(" ", "_")

        ### common abbreviation
        nicknames = {
            "acc": "accuracy",
            "auc": "roc_auc",
            "f1_score": "f1",
            "macro_f1": "f1_macro",
            "mcc": "matthews_corrcoef",
            "log_loss": "neg_log_loss",
            "cross_entropy": "neg_log_loss",
            "ce": "neg_log_loss",
            "bce": "neg_log_loss",
            "mse": "neg_mean_squared_error",
            "mean_squared_error": "neg_mean_squared_error",
            "mae": "neg_mean_absolute_error",
            "mean_absolute_error": "neg_mean_absolute_error",
            "mape": "neg_mean_absolute_percentage_error",
            "mean_absolute_percentage_error":
            "neg_mean_absolute_percentage_error",
            "rmse": "neg_root_mean_squared_error",
            "root_mean_squared_error": "neg_root_mean_squared_error",
            "qwk": "quadratic_weighted_kappa",
            "kappa": "cohen_kappa"
        }
        if scorer_name in nicknames:
            scorer_name = nicknames[scorer_name]

        self.metric_name = scorer_name

        # get the scorer
        ### kappa
        if scorer_name == "quadratic_weighted_kappa":
            return metrics.make_scorer(metrics.cohen_kappa_score,
                                       weights="quadratic",
                                       response_method="predict",
                                       greater_is_better=True)
        elif scorer_name == "cohen_kappa":
            return metrics.make_scorer(metrics.cohen_kappa_score,
                                       weights=None,
                                       response_method="predict",
                                       greater_is_better=True)
        ### others
        return metrics.get_scorer(scorer_name)

    def check_task(self) -> bool:
        """
        Uh....  this functions will do    
            1. some check for mis-using of method and types.    
            2. sparsing the task and target.    

        It will be nasty and full of dirty code. Make yourself at home.    
        

        Raises:
            ValueError: Check auc score only be used in binary classification

        Returns:
            bool: _description_
        """
        # Check auc score only be used in binary classification
        is_auc = self.metric_name == "roc_auc"
        self.is_binary = len(self.y.value_counts()) == 2
        if is_auc and not self.is_binary:
            # roc_auc only can be used on binary classification. Do not try ovr, ovo. forget them.
            raise ValueError(
                "auc only support binary classification, but more than 2 values are detected in y. Try 'f1_macro'"
            )

        # Sparse the metric
        scorer_kargs = {}
        for arguments in self.metric.__str__()[12:-1].split(", ")[1:]:
            if "(" in arguments:  # for roc auc scorer: response_method=('decision_function', 'predict_proba')
                scorer_kargs["response_method"] = "predict_proba"
            else:
                arg = arguments.split("=")
                if len(arg) == 2:
                    scorer_kargs[arg[0]] = arg[1]

        ### Response method
        self.metric_using_proba = scorer_kargs["response_method"].find(
            "_proba") != -1
        scorer_kargs.pop("response_method")

        ### Greater is better
        if 'greater_is_better' in scorer_kargs:
            self.metric_great_better = not scorer_kargs[
                "greater_is_better"] == "False"
            scorer_kargs.pop('greater_is_better')
        else:
            self.metric_great_better = True
        self.metric_limit = 1 if self.metric_great_better else 0

        ### pos_label for f1 socres
        if "pos_label" in scorer_kargs:
            scorer_kargs.pop("pos_label")

        self.scorer_kargs = scorer_kargs

        return True

    def plot(self):
        """
        Plot the learning process.
        """
        from plotly import io
        fig = optuna.visualization.plot_optimization_history(
            self.study, target_name=self.metric_name)
        fig.add_hline(y=self.default_performance,
                      line_dash="dot",
                      annotation_text="Default setting",
                      annotation_position="bottom right")
        io.show(fig)

    def _explainer(self, x: DataFrame) -> Explainer:
        print("Not implement")
        return None

    def shap_explain(self, x: DataFrame) -> DataFrame:
        """

        Args:
            x (DataFrame): the data to be explained which has shape (Samples, Features).
        
        Returns:
            explainer (shap.Explaination): Indexing by (Features, Output, Samples)
        """
        self.explainer = self._explainer(x)
        return self.explainer(x)

    def early_stop_callback(self, study, trial):
        if trial.value >= self.metric_limit:
            self.optuna_early_stop_counter -= 1
            if self.optuna_early_stop_counter <= 0:
                print("optuna evaluate value {} reachs {}'s maximun value {}".
                      format(study.best_value, self.metric_name,
                             self.metric_limit))
                study.stop()

    def detail(self):
        """
        show the experiment settings including:    
            1. models parameters searching range.    

        Returns:
            pandas.DataFrame
        """
        parms_range = DataFrame(self.parms_range()).drop(0)
        if self.best_model is not None:
            best_params = self.best_model.get_params()

            parms_range.loc[4] = Series(
                [best_params[param] for param in parms_range.columns],
                index=parms_range.columns)

        parms_range = parms_range.T.reset_index()
        parms_range.columns = [
            "parameter", "dtype", "lower_bound", "upper_bound", "result"
        ]

        parms_range.index = [" "] * parms_range.shape[0]

        name_holder = DataFrame({
            self.name(): {
                "parameter": None,
                "dtype": None,
                "lower_bound": None,
                "upper_bound": None,
                "result": None
            }
        }).T
        return concat([name_holder, parms_range], axis=0)



import matplotlib.pyplot as plt
from pandas import concat
import warnings
from sklearn.exceptions import ConvergenceWarning
from sklearn.model_selection import KFold
from sklearn.base import BaseEstimator
# Suppress only ConvergenceWarning
warnings.filterwarnings("ignore", category=ConvergenceWarning)


class SelectionPipeline(BaseEstimator):
    """
    The basic pipeline for selection methods. It includes 2 parts: Scoring and Choosing.
    The detail methods is to be determinded.

    """

    def __init__(self,
                 k: int = None,
                 z_importance_threshold: float = 1.,
                 n_cv: int = 1):
        """
        Initialize the selection pipeline.

        Args:
            k (int or None): select top k important feature. k = -1 means selecting all, k = None means selecting the feature that have standarized score > 1. Default = None
            z_importance_threshold (int, optional): The threshold to picking features. Defaults to 1.
        """
        self.k = k
        self.z_importance_threshold = z_importance_threshold
        self.n_cv = n_cv

    def _set_up(self):
        self.scores_ = None
        self.selected_score_ = None

    def Scoring(self, x, y=None):
        """
        The method to scores features is to be implemented.

        Args:
            x (pandas.DataFrame or a 2D array): The data to extract information.
            y (pandas.Series or a 1D array): The target label for methods. Defaults to None.

        Returns:
            pandas.Series or pandas.DataFrame: The score for each feature. Some elements may be empty.
        """
        scores = x.max().sort_values(ascending=False)  # top is better
        return scores

    def Select(self, scores):
        """
        pick features according to scores.
        if k == None, choose k such that:    
            z-scores = (scores - scores.mean())/scores.std()    
            k = # (z-scores > 1)
            
        Args:
            scores (pandas.Series): The score for each feature. Some elements may be empty.
        Returns:
            pandas.Series: The score for k selected features. May less than k.
        """
        # x should be a pd dataframe or a numpy array without missing value
        if self.k:
            # k not None
            selected_score = scores.head(self.k)
            selected_score = selected_score[selected_score != 0]
        else:
            # k is None
            z_scores = (scores - scores.mean()) / (scores.std() + 1e-4)
            selected_score = scores[z_scores > self.z_importance_threshold]
        selected_score.name = self.name
        return selected_score

    def fit(self, x, y):
        """
        sklearn api

        Args:
            x (pandas.DataFrame or a 2D array): The data to extract information.
            y (pandas.Series or a 1D array): The target label for methods.
        """
        self._set_up()

        if self.n_cv > 1:
            scores = []
            for train_idx, valid_idx in KFold(n_splits=self.n_cv,
                                              shuffle=True,
                                              random_state=20250915).split([
                                                  i for i in range(x.shape[0])
                                              ]):
                scores.append(
                    self.Scoring(x.iloc[train_idx], y.iloc[train_idx]))
            self.scores_ = (concat(scores, axis=1).sum(axis=1) /
                            self.n_cv).sort_values(ascending=False)
        else:
            self.scores_ = self.Scoring(x.copy(), y.copy())
        self.selected_score_ = self.Select(self.scores_.copy())
        return self

    def transform(self, x):
        return x[self.selected_score_.index]

    def fit_transform(self, x, y):
        self.fit(x, y)
        return self.transform(x)

    def what_matters(self):
        return self.selected_score_

    def Plotting(self):
        """
        plot hist graph of selectied feature importance
        """
        plt.rcParams['font.family'] = 'Arial'
        plt.rcParams['font.size'] = 14
        fig, ax = plt.subplots(1, 1)
        ax.bar(self.selected_score_.index, self.selected_score_)
        for label in ax.get_xticklabels(which='major'):
            label.set(rotation=45, horizontalalignment='right')
        ax.set_title(self.name + " score")
        ax.set_ylabel("Importance")

        plt.show()

    def reference(self) -> dict[str, str]:
        """
        This function will return reference of this method in python dict.    
        If you want to access it in PineBioML api document, then click on the    >Expand source code     

        Returns:
            dict[str, str]: a dict of reference.
        """
        refers = {
            "sklearn publication":
            "https://dl.acm.org/doi/10.5555/1953048.2078195"
        }

        return refers

    def Report(self):
        """
        Give diagnose of selection.
        """
        #pass



import sklearn.preprocessing as skprpr
from pandas import DataFrame
from sklearn.base import BaseEstimator


class RemoveDummy(BaseEstimator):
    """ 
    Remove dummy features. Dummy features are those with a constant value.
    """

    def __init__(self):
        pass

    def fit(self, x, y=None):
        """
        Finding the dummy features.

        Args:
            x (pandas.DataFrame or a 2D array): The data to normalize.
            y (pandas.Series or a 1D array): A placeholder only. Normalizer do nothing to y.

        Returns:
            remove: self after fitting.
        """

        self.to_drop_ = x.columns[x.nunique() <= 1]
        return self

    def transform(self, x):
        """
        Drop the dummy features. Only activates after "fit" was called.

        Args:
            x (pandas.DataFrame or a 2D array): The data to drop dummy features.

        Returns:
            pandas.DataFrame or a 2D array: Cleared x.
        """
        x_dropped = x.drop(self.to_drop_, axis=1)

        return x_dropped

    def fit_transform(self, x, y=None):
        """
        A functional stack of "fit" and "transform".

        Args:
            x (pandas.DataFrame or a 2D array): The data to normalize.
            y (pandas.Series or a 1D array): A placeholder only. Normalizer do nothing to y.

        Returns:
            pandas.DataFrame or a 2D array: Cleared x.
        """
        self.fit(x, y)
        x_dropped = self.transform(x)
        return x_dropped


class Normalizer(BaseEstimator):
    """ 
    A wrapper of sklearn normalizers. This will conserve pandas features.    
    method be one of ["StandardScaler", "RobustScaler", "MinMaxScaler", "Normalizer", "PowerTransformer"]    
    let x with shape [n, d] where n is the sample size, d is the number of features.    
        StandardScaler(x) = (x - x.mean(n)) / x.std(n)    
        RobustScaler(x) = (x - x.median(n)) / (x.quartile3(n) - x.quartile1(n))    
        MinMaxScaler(x) = (x - x.min(n)) / (x.max(n) - x.min(n))    
        Normalizer(x) = x / x.norm(d)
        PowerTransformer(x) = yeo-johnson transform(x)
    """

    def __init__(self, center=True, scale=True, method="RobustScaler"):
        """
        Args:
            center (bool, optional): Whether to centralize data. Default to True.
            scale (bool, optional): Whether to scaling data after centralized. Default to True.
            method (str, optional): the way to normalize data. Be one of ["StandardScaler", "RobustScaler", "MinMaxScaler", "Normalizer", "PowerTransformer"]
        """
        self.center = center
        self.scale = scale
        self.method = method

    def _set_up(self):
        kernels = {
            "StandardScaler":
            skprpr.StandardScaler(with_mean=self.center, with_std=self.scale),
            "RobustScaler":
            skprpr.RobustScaler(with_centering=self.center,
                                with_scaling=self.scale),
            "MinMaxScaler":
            skprpr.MinMaxScaler(),
            "Normalizer":
            skprpr.Normalizer(),
            "PowerTransformer":
            skprpr.PowerTransformer()
        }
        self.kernel_ = kernels[self.method]

    def fit(self, x, y=None):
        """
        Computing and Recording the mean and std of X for prediction.    

        Args:
            x (pandas.DataFrame or a 2D array): The data to normalize.    
            y (pandas.Series or a 1D array): A placeholder only. Normalizer do nothing to y.    

        Returns:
            Normalizer: self after fitting.
        """
        self._set_up()
        self.kernel_.fit(x, y)

        return self

    def transform(self, x):
        """
        Transform input x. Only activates after "fit" was called.

        Args:
            x (pandas.DataFrame or a 2D array): The data to normalize.    
            y (pandas.Series or a 1D array): A placeholder only. Normalizer do nothing to y.    

        Returns:
            pandas.DataFrame or a 2D array: Normalized x.
            pandas.Series or a 1D array: Same as input y.
        """
        x_normalized = DataFrame(self.kernel_.transform(x),
                                 index=x.index,
                                 columns=x.columns)

        return x_normalized

    def fit_transform(self, x, y=None):
        """
        A functional stack of "fit" and "transform".

        Args:
            x (pandas.DataFrame or a 2D array): The data to normalize.
            y (pandas.Series or a 1D array): A placeholder only. Normalizer do nothing to y.

        Returns:
            pandas.DataFrame or a 2D array: Normalized x.
        """
        self.fit(x, y)
        x_normalized = self.transform(x)
        return x_normalized

    def inverse_transform(self, x):
        """
        The inverse transform of normalize.

        Args:
            x (pandas.DataFrame or a 2D array): The data to revert original scale.
            y (pandas.Series or a 1D array): A placeholder only. Normalizer do nothing to y.

        Returns:
            pandas.DataFrame or a 2D array: x in original scale.
        """

        x_original = DataFrame(self.kernel_.inverse_transform(x),
                               index=x.index,
                               columns=x.columns)

        return x_original


class Pass(BaseEstimator):
    """ 
    Do nothing.
    """

    def __init__(self):
        pass

    def fit(self, x, y=None):
        """
        Do nothing.    

        Args:
            x (pandas.DataFrame or a 2D array): The data to normalize.    
            y (pandas.Series or a 1D array): A placeholder only. Normalizer do nothing to y.    

        Returns:
            Pass: No, do nothing.
        """
        self.do_nothing_ = True

        return self

    def transform(self, x):
        """
        Do nothing.

        Args:
            x (pandas.DataFrame or a 2D array): The data to normalize.    

        Returns:
            pandas.DataFrame or a 2D array: Normalized x.
        """

        return x

    def fit_transform(self, x, y=None):
        """
        Do nothing.

        Args:
            x (pandas.DataFrame or a 2D array): The data to normalize.
            y (pandas.Series or a 1D array): A placeholder only. Normalizer do nothing to y.

        Returns:
            pandas.DataFrame or a 2D array: x.
        """
        self.fit(x, y)
        x_normalized = self.transform(x)
        return x_normalized

    def inverse_transform(self, x):
        """
        Do nothing.

        Args:
            x (pandas.DataFrame or a 2D array): The data to revert original scale.

        Returns:
            pandas.DataFrame or a 2D array: x.
        """

        return x



from sklearn.impute import KNNImputer, SimpleImputer
import pandas as pd
from . import Normalizer
from typing import Literal
from sklearn.base import BaseEstimator


class imputer(BaseEstimator):
    """
    To impute missing value. Include 4 parts:

    1. Deleting the features with too high missing value ratio.    
    2. Normalize input. Some imputed methods are sensitive to scale.    
    3. impute. method to be determinded.    
    4. inverse normalize.    
    """

    def __init__(self, threshold=0.333, center=True, scale=True):
        """__init__ _summary_

        Args:
            threshold (float): float from (0, 1]. If missing value rate of a feature is higher than threshold, it will be deleted. Defaults to 0.333
            center (bool, optional): To centralize data or not. Defaults to True.
            scale (bool, optional): To scale data or not. Defaults to True.

        Raises:
            ValueError: missing value threshold must be a float in (0, 1]
        """
        self.threshold = threshold
        self.center = center
        self.scale = scale

    def _set_up(self):
        self.normalizer = Normalizer(center=self.center, scale=self.scale)

    def fit(self, x, y=None, sample_weight=None):
        """
        Fit x.
    
        Args:
            x (pandas.DataFrame or a 2D array): The data to extract information.
            y (pandas.Series or a 1D array): The target label for methods. Defaults to None.

        Returns:
            inputer: fitted self
        """
        self._set_up()

        # drop too empty features
        self.not_too_empty = x.isna().mean() <= self.threshold
        x = x.loc[:, self.not_too_empty]  # keep those who not too empty

        # normalize
        self.normalizer.fit(x, y)
        x = self.normalizer.transform(x)

        # call the kernel
        self.kernel.fit(x)
        self.fitted_ = True
        return self

    def transform(self, x):
        """
        Transform x

        Args:
            x (pandas.DataFrame or a 2D array): The data to extract information.
            y (pandas.Series or a 1D array): The target label for methods. Defaults to None.

        Returns:
            pandas.DataFrame or a 2D array: imputed x
        """
        if not self.fitted_:
            raise "please call fit before calling transform."
        # drop too empty features
        x = x.loc[:, self.not_too_empty]  # keep those who not too empty

        columns = x.columns
        idx = x.index

        # normalize
        x = self.normalizer.transform(x)

        # call the kernel
        x = self.kernel.transform(x)

        # rebuild the dataframe from numpy array returns
        x = pd.DataFrame(x, columns=columns, index=idx)

        # inverse normalize
        x = self.normalizer.inverse_transform(x)
        return x

    def fit_transform(self, x, y=None, sample_weight=None):
        """
        A functional stack of fit and transform.

        Args:
            x (pandas.DataFrame or a 2D array): The data to extract information.
            y (pandas.Series or a 1D array): The target label for methods. Defaults to None.

        Returns:
            pandas.DataFrame or a 2D array: imputed x
        """
        self.fit(x, y)
        x = self.transform(x)
        return x

    def report(self):
        state = {
            "total features": len(self.not_too_empty),
            "drop features":
            len(self.not_too_empty) - self.not_too_empty.sum()
        }
        return state


class knn_imputer(imputer):
    """
    Using K nearest neighbor to impute missing value

    """

    def __init__(self, threshold=0.333, n_neighbor=5):
        """

        Args:
            n_neighbor (int, optional): Number of nearest neighbor to use. Defaults to 5.
        """
        super().__init__(threshold)
        self.n_neighbor = n_neighbor

    def _set_up(self):
        self.kernel = KNNImputer(n_neighbors=self.n_neighbor)
        self.normalizer = Normalizer(center=self.center, scale=self.scale)


class simple_imputer(imputer):
    """
    Using a constant value (0, mean, median... etc.) to impute all missing value in a feature.
    """

    def __init__(self, threshold=0.333, strategy="median"):
        """

        Args:
            strategy (str, optional): The strategy to impute. One of {"mean", "median", "constant"}. Defaults to "median".
        Todo: 
            1. filling value for constant mode.
        """
        super().__init__(threshold, center=False, scale=False)
        self.strategy = strategy

    def _set_up(self):
        self.kernel = SimpleImputer(strategy=self.strategy)
        self.normalizer = Normalizer(center=self.center, scale=self.scale)


class iterative_imputer(imputer):
    """
    Using iterative imputer to impute missing value.
    """

    def __init__(self,
                 threshold=0.333,
                 max_iter=10,
                 estimator: Literal["RandomForest",
                                    "LightGBM"] = "RandomForest"):
        """

        Args:
            max_iter (int, optional): The maximum number of imputation iteration. Defaults to 10.
        """
        super().__init__(threshold)
        self.max_iter = max_iter
        self.estimator = estimator

    def _set_up(self):

        from sklearn.experimental import enable_iterative_imputer
        from sklearn.impute import IterativeImputer

        if self.estimator == "RandomForest":
            from sklearn.ensemble import RandomForestRegressor

            self.kernel = IterativeImputer(estimator=RandomForestRegressor(
                n_estimators=256,
                n_jobs=-1,
                min_samples_leaf=3,
                max_features="sqrt",
                bootstrap=True,
                max_samples=0.7,
                random_state=143),
                                           max_iter=self.max_iter,
                                           imputation_order="random",
                                           verbose=1)
        elif self.estimator == "LightGBM":
            from lightgbm import LGBMRegressor

            self.kernel = IterativeImputer(estimator=LGBMRegressor(
                n_estimators=100,
                n_jobs=None,
                min_child_samples=3,
                verbosity=-1,
                reg_lambda=0.1,
                random_state=143),
                                           max_iter=self.max_iter,
                                           imputation_order="random",
                                           verbose=1)
        else:
            raise ValueError(
                "estimator must be one of [RandomForest, LightGBM]")



from . import SelectionPipeline

import pandas as pd
import numpy as np
from tqdm import tqdm
import time

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.cm import ScalarMappable

from joblib import Parallel, parallel_config, delayed

from sklearn.utils.class_weight import compute_sample_weight
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model import Lasso, LassoLars, LogisticRegression
from sklearn.svm import LinearSVC


class Lasso_selection(SelectionPipeline):
    """
    Using Lasso (L1 penalty) regression as scoring method. L1 penalty will force feature weights to be zeros.    
    As the penalty increases, more and more regression coefficients vanish and the ones coresponding to important variables will remain.    

    The Lasso_selection is base on [Lasso Lars](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLars.html).    

    Currently, this do not support Lasso + logistic regression, so:    
        Binary classification problem will be regarded as a regression to {0, 1}.    
        Multi-class classification problem will be devided into ovr (one vs rest) classification and the score of features over ovr will be combined by average weighted class weight.    

    ~~Lasso_selection will use grid search to find out when all weights vanish.    ~~
    """

    def __init__(self,
                 k=None,
                 z_importance_threshold=1.,
                 unbalanced=True,
                 n_cv=5):
        """
        Args:
            unbalanced (bool, optional): False to imply class weight to samples. Defaults to True.
        """
        super().__init__(k=k,
                         z_importance_threshold=z_importance_threshold,
                         n_cv=n_cv)

        # parameters
        self.regression = True
        self.unbalanced = unbalanced
        self.name = "LassoLars"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoLars.html"
        refer[
            " publication Lasso 1"] = "https://projecteuclid.org/journals/annals-of-statistics/volume-37/issue-5A/High-dimensional-variable-selection/10.1214/08-AOS646.full"
        refer[
            " publication Lasso 2"] = "https://www.tandfonline.com/doi/abs/10.1198/016214506000000735?casa_token=5HDhtyCfh40AAAAA:4NxSU97CZubZVpReaQNsSBpqA10_xNhspTQobPnb_z2YXe3Wf-HBHV8OygbqUkmJQPt2Jmp7ZlJPWd0"
        refer[" publication Lars"] = "https://arxiv.org/pdf/math/0406456"

        return refer

    def create_kernel(self, C=1e-2):
        """
        Create diffirent kernel according to opjective.

        Args:
            C (float): The coefficient to L1 penalty.

        Returns:
            sklearn.linearmodel: a kernel of sklearn linearmodel
        
        TODO:
            1. auto C tuner or adapter.
            
        """

        return LassoLars(alpha=C, random_state=142)

    def Scoring(self, x, y=None):
        """
        Using Lasso Lars regression as scoring method.
                
        """

        x_train = x.copy()
        y_train = y.copy()

        self.y_encoder = OneHotEncoder(sparse_output=False)
        self.y_encoder.fit(y.to_numpy().reshape(-1, 1))
        y_train = self.y_encoder.transform(y_train.to_numpy().reshape(-1, 1))

        kernel = self.create_kernel()

        ### TODO: Sample weight for Lars.
        if y_train.shape[1] == 2:
            # Binary classification
            task_set = [1]
        else:
            # multi-class classification
            ### TODO: Multinomial classfication for Lars.
            task_set = range(y_train.shape[1])

        result = []
        for i in task_set:
            # TODO: cross validation, alpha interpolation
            y = y_train[:, i]
            y = (y - y.mean()) / y.std()
            kernel.fit(x_train, y)

            coef = np.clip(kernel.coef_path_.T, -1, 1)
            alpha = kernel.alphas_
            col = kernel.feature_names_in_

            result.append(pd.DataFrame(coef, columns=col, index=alpha))

        var_dropout_alpha = []
        for s in result:
            # the right end side(alpha) of the range that variable has a non-zero lars coefficient.
            tmp = (s != 0).idxmax(axis=0).replace(s.index[0], 0)**2

            # normalizing each task.
            tmp = tmp / tmp.max()
            var_dropout_alpha.append(tmp)

        # collect the results
        self.result = result
        scores = np.sqrt(sum(var_dropout_alpha)).sort_values(ascending=False)
        scores.name = self.name
        return scores

    def Plotting(self):
        super().Plotting()

        global_selected = self.selected_score_.index
        for i_th in range(len(self.result)):
            s = self.result[-i_th - 1]

            # get the features ever alive.
            alive_col = s.columns[(s != 0).any(axis=0)]

            # show the label of the top-5 long-lived features
            noting_col = (s != 0).idxmax(axis=0).replace(
                s.index[0], 0).sort_values().tail(5).index

            plt_cmap = plt.get_cmap("Blues")
            plt_norm = plt.Normalize(vmin=0, vmax=s.index[1])
            fig, ax = plt.subplots(layout='constrained')

            for col in alive_col:
                var_coef = s[col]
                self.plot_coef_1var(
                    alpha=var_coef.index,
                    coef=var_coef.values,
                    global_selected=col in global_selected,
                    coef_name=col if col in noting_col else None,
                    cmap=plt_cmap,
                    norm=plt_norm,
                )

            plt.axhline(y=0, color="grey", linestyle="--")
            plt.xlabel("alpha")
            plt.ylabel("coefficient")
            plt.title("class-{} Lasso".format(
                self.y_encoder.categories_[0][i_th]))

            scalar_mappable = ScalarMappable(norm=plt_norm, cmap=plt_cmap)
            fig.colorbar(scalar_mappable,
                         ax=ax,
                         orientation='vertical',
                         label='Variable dropout alpha')
            plt.xscale('log')
            legend_elements = [
                Line2D([0], [0],
                       color="b",
                       label="Globally important",
                       linestyle="solid"),
                Line2D([0], [0],
                       color="b",
                       label="Partially important",
                       linestyle="--"),
            ]
            ax.legend(handles=legend_elements, loc="upper right")
            plt.show()

    def plot_coef_1var(self, alpha, coef, global_selected, coef_name, cmap,
                       norm):
        # plot the coef curve
        linestyle = "solid" if global_selected else "--"

        alive = coef != 0
        alive[alive.argmax() - 1] = True

        alpha = alpha[alive]
        coef = coef[alive]

        plt.plot(alpha,
                 coef,
                 label=coef_name,
                 c=cmap(norm(alpha.max())),
                 linestyle=linestyle)

        # plot the annotate
        if coef_name is not None:
            annotate_idx = np.abs(coef).argmax()
            annotate_coor = (alpha[annotate_idx], coef[annotate_idx])
            text_coor = (alpha[annotate_idx], coef[annotate_idx])
            plt.annotate(
                coef_name,
                xy=annotate_coor,
                xytext=text_coor,
            )


class multi_Lasso_selection(SelectionPipeline):
    """
    A stack of Lasso_selection. Because of collinearity, if there are a batch of featres with high corelation, only one of them will remain.    
    That leads to diffirent behavior between select k features in a time and select k//n features in n times.    
    """

    def __init__(self, k=None, z_importance_threshold=1., n=4, n_cv=5):
        """
        Args:
            n (int , optional): How many times should lasso be performed.    
            objective (str, optional): one of {"Regression", "BinaryClassification"}
        """
        super().__init__(k=k,
                         z_importance_threshold=z_importance_threshold,
                         n_cv=n_cv)
        self.name = "multi_Lasso"
        self.backend = Lasso_selection  #Lasso_bisection_selection
        self.n = n
        self.n_cv = 1  # multi lasso calls lasso which already applied cv

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            " Warning"] = "We do not have a reference and this method's effectivity has not been proven yet."
        return refer

    def Scoring(self, x, y):
        result = []
        if self.k == -1 or self.k is None:
            self.k = min(x.shape[0], x.shape[1]) // 2
        batch_size = self.k // self.n + 1
        loop_lim = x.shape[1] // batch_size

        num_selected = 0
        counter = 0
        while (num_selected < self.k):
            kernel = self.backend(k=batch_size).fit(x, y)
            result.append(kernel.selected_score_)
            batch_selected = result[-1].index
            x = x.drop(batch_selected, axis=1)
            num_selected += len(batch_selected)
            if x.shape[1] == 0:
                break

            counter += 1
            if counter > loop_lim:
                break

        result = pd.concat(result).sort_values(ascending=False)
        #result = result - result.min()
        result.name = self.name

        return result


class SVM_selection(SelectionPipeline):
    """
    Using the support vector of linear support vector classifier as scoring method.    

    SVM_selection is scale sensitive in result.    

    <<Feature Ranking Using Linear SVM>> section 3.2

    """

    def __init__(self, k=None, z_importance_threshold=1., n_cv=5):
        super().__init__(k=k,
                         z_importance_threshold=z_importance_threshold,
                         n_cv=n_cv)
        self.kernel = LinearSVC(dual="auto",
                                class_weight="balanced",
                                random_state=142)
        self.name = "SVM"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html"
        refer[
            " publication - section 3.2"] = "https://www.csie.ntu.edu.tw/~cjlin/papers/causality.pdf"
        return refer

    def Scoring(self, x, y=None):
        """
        Using the support vector of linear support vector classifier as scoring method.

        Returns:
            pandas.Series or pandas.DataFrame: The score for each feature. Some elements may be empty.
        """
        self.kernel.fit(x, y)
        svm_weights = np.abs(self.kernel.coef_).sum(axis=0)
        svm_weights /= svm_weights.sum()

        scores = pd.Series(svm_weights, index=x.columns,
                           name=self.name).sort_values(ascending=False)
        return scores


class DT_selection(SelectionPipeline):
    """
    A child class of SelectionPipeline.

    Using Decision stump (a single Decision tree) to scoring features.    
    What we do here is:    
        1. normalize data    
        2. transform data into frequency domain by binning through a certain column and applying value_counts    
        3. estimate entropy    
        4. compute c4.5    
    """

    def __init__(self,
                 k=None,
                 z_importance_threshold=1.,
                 bins=10,
                 q=0.05,
                 strategy="c45",
                 n_cv=5):
        """
        Args:
            bins (int, optional): Bins to esimate data distribution entropy. Defaults to 10.
            q (float, optional): Clip data values out of [q, 1-q] percentile to reduce the affect of outliers while estimate entropy. Defaults to 0.05.
            strategy (str, optional): One of {"gini", "c45"}. The strategy to build decision tree. Defaults to "c45".
        """
        super().__init__(k=k,
                         z_importance_threshold=z_importance_threshold,
                         n_cv=n_cv)
        self.bins = bins - 1
        self.q = q
        self.strategy = strategy
        self.name = "DT_score_" + self.strategy

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[self.name() +
              " document"] = "PineBioML.selection.classification.DT_selection"
        refer[
            " publication"] = "Quinlan, J. R. C4.5: Programs for Machine Learning. Morgan Kaufmann Publishers, 1993."
        return refer

    def Scoring(self, x, y=None):
        """
        Using Decision stump (a single Decision tree) to scoring features. Though, single layer stump is equivalent to compare the id3/c4.5 score directly.

        Returns:
            pandas.Series or pandas.DataFrame: The score for each feature. Some elements may be empty.
        """
        self.y_encoder = OneHotEncoder(sparse_output=False)
        y = self.y_encoder.fit_transform(y.to_numpy().reshape(
            -1, 1)).argmax(axis=-1)

        if self.strategy == 'c45':
            score = self.best_splits_c45_batch_optimized(x, y)
        elif self.strategy == 'gini':
            score = self.best_splits_gini_batch_optimized(x, y)

        scores = pd.Series(score, index=x.columns,
                           name=self.name).sort_values(ascending=False)
        return scores

    def compute_gain_ratio(self, y_sorted, n_classes, n, eps=1e-12):
        """
        Compute best Gain Ratio for one sorted feature
        """
        # one-hot encode
        one_hot = np.eye(n_classes)[y_sorted]

        # cumulative counts (exclude last split)
        left_counts = np.cumsum(one_hot, axis=0)[:-1]
        total_counts = left_counts[-1]
        right_counts = total_counts - left_counts

        # sample counts
        left_totals = np.arange(1, n)
        right_totals = n - left_totals

        # ---- entropy computation ----
        def entropy(counts, totals):
            probs = counts / totals[:, None]
            probs = np.clip(probs, eps, 1.0)
            return -np.sum(probs * np.log2(probs), axis=1)

        # parent entropy
        parent_probs = total_counts / n
        parent_probs = np.clip(parent_probs, eps, 1.0)
        parent_entropy = -np.sum(parent_probs * np.log2(parent_probs))

        # left / right entropy
        entropy_left = entropy(left_counts, left_totals)
        entropy_right = entropy(right_counts, right_totals)

        # information gain
        info_gain = parent_entropy - ((left_totals / n) * entropy_left +
                                      (right_totals / n) * entropy_right)

        # split information
        p_left = left_totals / n
        p_right = right_totals / n
        split_info = -(p_left * np.log2(np.clip(p_left, eps, 1.0)) +
                       p_right * np.log2(np.clip(p_right, eps, 1.0)))

        # gain ratio
        gain_ratio = info_gain / np.clip(split_info, eps, None)

        # C4.5 chooses the **maximum** gain ratio
        return np.max(gain_ratio)

    def compute_gini(self, y_sorted, n_classes, n):
        one_hot = np.eye(n_classes)[y_sorted]
        left_counts = np.cumsum(one_hot, axis=0)[:-1]

        # right counts
        total_counts = left_counts[-1]
        right_counts = total_counts - left_counts

        # total counts
        left_totals = np.arange(1, n)
        right_totals = n - left_totals

        # vectorizing left & right Gini computing
        gini_left = 1.0 - np.sum(
            (left_counts / left_totals[:, np.newaxis])**2, axis=1)
        gini_right = 1.0 - np.sum(
            (right_counts / right_totals[:, np.newaxis])**2, axis=1)

        # weighted Gini
        weighted_gini = (left_totals * gini_left +
                         right_totals * gini_right) / n
        return np.min(weighted_gini)

    def best_splits_gini_batch_optimized(self, X, Y):
        """
        compute gini in batch
        
        Args:
            X (pandas.DataFrame): (n_samples, n_features) in numerical dtypes.
            y (pandas.Series): (n_samples,) in int.
        
        return:
            best_ginis: (n_features,)
            best_thresholds: (n_features,) placeholder, we don't need it here
        """
        x = X  #.to_numpy()
        y = Y  #.to_numpy().astype(np.int8)
        n, n_features = X.shape
        n_classes = y.max() + 1

        # sorting all the features
        sort_idx = np.argsort(x, axis=0)

        #X_sorted = np.take_along_axis(X, sort_idx, axis=0)
        y_sorted_all = np.take_along_axis(np.tile(y[:, None], (1, n_features)),
                                          sort_idx,
                                          axis=0)

        best_ginis = np.ones(n_features)

        with parallel_config(backend='loky', n_jobs=-1):
            best_ginis = Parallel()(
                delayed(self.compute_gini)(y_sorted_all[:, i], n_classes, n)
                for i in tqdm(np.arange(n_features)))

        return best_ginis

    def best_splits_c45_batch_optimized(self, X, Y):
        """
        compute C4.5 gain ratio in batch
        
        Args:
            X (pandas.DataFrame): (n_samples, n_features)
            Y (pandas.Series): (n_samples,) int labels
        
        return:
            best_gain_ratios: (n_features,)
        """
        x = X
        y = Y
        n, n_features = X.shape
        n_classes = y.max() + 1

        # sort indices per feature
        sort_idx = np.argsort(x, axis=0)

        # reorder y for each feature
        y_sorted_all = np.take_along_axis(np.tile(y[:, None], (1, n_features)),
                                          sort_idx,
                                          axis=0)

        with parallel_config(backend='loky', n_jobs=-1):
            best_gain_ratios = Parallel()(
                delayed(self.compute_gain_ratio)(y_sorted_all[:,
                                                              i], n_classes, n)
                for i in tqdm(range(n_features)))

        return np.array(best_gain_ratios)


class RF_selection(SelectionPipeline):
    """
    Using random forest to scoring features by gini/entropy gain.    
    We do not provide permutation importance(VI, variable importance) here.

    """

    def __init__(self,
                 k=None,
                 z_importance_threshold=1.,
                 unbalanced=True,
                 strategy="gini",
                 n_cv=5):
        """
        Args:
            strategy (str, optional): Scoring strategy, one of {"gini", "entropy"}. Defaults to "gini".
            unbalanced (bool, optional): True to imply class weight to samples. Defaults to True.
        """
        from sklearn.ensemble import RandomForestClassifier
        super().__init__(k=k,
                         z_importance_threshold=z_importance_threshold,
                         n_cv=n_cv)
        self.strategy = strategy
        if unbalanced:
            class_weight = "balanced"
        else:
            class_weight = None

        self.kernel = RandomForestClassifier(n_estimators=100,
                                             n_jobs=-1,
                                             max_samples=0.7,
                                             class_weight=class_weight,
                                             criterion=strategy,
                                             verbose=0,
                                             random_state=142,
                                             min_samples_leaf=2)

        self.name = "RandomForest_" + self.strategy

    def reference(self) -> dict[str, str]:

        refer = super().reference()
        refer[self.name() + " document"] = ""
        refer[
            " publication cons"] = "https://link.springer.com/article/10.1186/1471-2105-8-25"
        refer[
            " publication pros 1"] = "https://link.springer.com/article/10.1186/1471-2105-10-213"
        refer[
            " publication pros 2"] = "https://www.sciencedirect.com/science/article/pii/S0167947307003076"
        refer[
            " publication survey"] = "https://www.cs.cmu.edu/~qyj/papersA08/11-rfbook.pdf"

        return refer

    def Scoring(self, x, y=None):
        """
        Using random forest to scoring (gini impurity / entropy) features.

        Returns:
            pandas.Series or pandas.DataFrame: The score for each feature. Some elements may be empty.
        """
        self.kernel.n_estimators = round(2 * np.sqrt(x.shape[1]) *
                                         np.log(x.shape[1]))
        with parallel_config(backend='loky'):
            self.kernel.fit(x, y)
        score = self.kernel.feature_importances_
        scores = pd.Series(score, index=x.columns,
                           name=self.name).sort_values(ascending=False)
        return scores


class XGboost_selection(SelectionPipeline):
    """
    Using XGboost to scoring (gini impurity / entropy) features.

    Warning: If data is too easy, boosting methods is difficult to give score to all features.
    """

    def __init__(self,
                 k=None,
                 z_importance_threshold=1.,
                 unbalanced=True,
                 n_cv=5):
        """
        Args:
            unbalanced (bool, optional): True to imply class weight to samples. Defaults to False.
        """
        super().__init__(k=k,
                         z_importance_threshold=z_importance_threshold,
                         n_cv=n_cv)
        self.unbalanced = unbalanced

        from xgboost import XGBClassifier

        self.kernel = XGBClassifier(random_state=142, subsample=0.7)
        self.name = "XGboost"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://xgboost.readthedocs.io/en/stable/python/python_api.html#xgboost.XGBRegressor.feature_importances_"
        return refer

    def Scoring(self, x, y=None):
        """
        Using XGboost to scoring (gini impurity / entropy) features.

        Returns:
            pandas.Series or pandas.DataFrame: The score for each feature. Some elements may be empty.
        """
        if self.unbalanced:
            sample_weight = compute_sample_weight(class_weight="balanced", y=y)
        else:
            sample_weight = np.ones(len(y))

        y = OneHotEncoder(sparse_output=False).fit_transform(
            y.to_numpy().reshape(-1, 1))
        self.kernel.fit(x, y, sample_weight=sample_weight)
        score = self.kernel.feature_importances_
        scores = pd.Series(score, index=x.columns,
                           name=self.name).sort_values(ascending=False)
        return scores


class Lightgbm_selection(SelectionPipeline):
    """
    Using Lightgbm to scoring (gini impurity / entropy) features. 

    Warning: If data is too easy, boosting methods is difficult to give score to all features.
    """

    def __init__(self,
                 k=None,
                 z_importance_threshold=1.,
                 unbalanced=True,
                 n_cv=5):
        """
        Args:
            unbalanced (bool, optional): True to imply class weight to samples. Defaults to False.
        """
        super().__init__(k=k,
                         z_importance_threshold=z_importance_threshold,
                         n_cv=n_cv)
        self.unbalanced = unbalanced
        from lightgbm import LGBMClassifier

        self.kernel = LGBMClassifier(learning_rate=0.01,
                                     random_state=142,
                                     subsample=0.7,
                                     subsample_freq=1,
                                     verbosity=-1)
        self.name = "Lightgbm"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://lightgbm.readthedocs.io/en/latest/Parameters.html#saved_feature_importance_type"
        return refer

    def Scoring(self, x, y=None):
        """
        Using Lightgbm to scoring (gini impurity / entropy) features.

        Returns:
            pandas.Series or pandas.DataFrame: The score for each feature. Some elements may be empty.
        """
        if self.unbalanced:
            sample_weight = compute_sample_weight(class_weight="balanced", y=y)
        else:
            sample_weight = np.ones(len(y))

        self.kernel.fit(x, y, sample_weight=sample_weight)
        score = self.kernel.feature_importances_
        scores = pd.Series(score, index=x.columns,
                           name=self.name).sort_values(ascending=False)
        return scores


class AdaBoost_selection(SelectionPipeline):
    """
    Using AdaBoost to scoring (gini impurity / entropy) features.

    Warning: If data is too easy, boosting methods is difficult to give score to all features.
    """

    def __init__(self,
                 k=None,
                 z_importance_threshold=1.,
                 unbalanced=True,
                 n_iter=128,
                 learning_rate=0.01,
                 n_cv=5):
        """
        Args:
            n_iter (int, optional): Number of trees also number of iteration to boost. Defaults to 64.
            learning_rate (float, optional): boosting learning rate. Defaults to 0.01.
            unbalanced (bool, optional): True to imply class weight to samples. Defaults to False.
        """
        from sklearn.ensemble import AdaBoostClassifier
        super().__init__(k=k,
                         z_importance_threshold=z_importance_threshold,
                         n_cv=n_cv)
        self.unbalanced = unbalanced
        self.kernel = AdaBoostClassifier(
            n_estimators=n_iter,
            learning_rate=learning_rate,
            random_state=142,
        )
        self.name = "AdaBoost" + str(n_iter)

    def reference(self) -> dict[str, str]:

        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostClassifier.html#sklearn.ensemble.AdaBoostClassifier.feature_importances_"
        return refer

    def Scoring(self, x, y=None):
        """
        Using AdaBoost to scoring (gini impurity / entropy) features.

        Returns:
            pandas.Series or pandas.DataFrame: The score for each feature. Some elements may be empty.
        """
        print("I don't have a progress bar but I am running")
        if self.unbalanced:
            sample_weight = compute_sample_weight(class_weight="balanced", y=y)
        else:
            sample_weight = np.ones(len(y))
        self.kernel.fit(x, y, sample_weight)
        score = self.kernel.feature_importances_
        scores = pd.Series(score, index=x.columns,
                           name=self.name).sort_values(ascending=False)
        return scores


class ensemble_selector(SelectionPipeline):
    """
    A functional stack of diffirent methods.    
    What we do here is:    
        1. calculate feature importance in different methods.    
        2. standardize the scores and then averaging through methods.    
        3. If z_importance_threshold not None, then all features with averaging score higher than z_importance_threshold will be selected.    
           else top k feature with averaging score will be selected.    
    """

    def __init__(self,
                 k: int = None,
                 z_importance_threshold: float = 1.,
                 n_cv=5):

        super().__init__(k=k,
                         z_importance_threshold=z_importance_threshold,
                         n_cv=n_cv)
        self.name = "ensemble"
        self.kernels = {
            "c45": DT_selection(k=k, strategy="c45", n_cv=n_cv),
            "RF_gini": RF_selection(k=k, strategy="gini", n_cv=n_cv),
            "Lasso": Lasso_selection(k=k, n_cv=n_cv),
            "multi_Lasso": multi_Lasso_selection(k=k, n_cv=n_cv),
            "SVM": SVM_selection(k=k, n_cv=n_cv),

            #"AdaBoost": AdaBoost_selection(k=k),
            #"XGboost": XGboost_selection(k=k),
            #"Lightgbm": Lightgbm_selection(k=k)
        }
        self.n_cv = 1

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " WARNING"] = "This method has no reference yet. That means the effectivity has not been proven yet. It somehow works in experience."
        return refer

    def Scoring(self, x, y=None):
        results = []
        for method in self.kernels:
            print("Using ", method, " to score.")
            start_time = time.time()

            self.kernels[method].fit(x, y)

            results.append(self.kernels[method].selected_score_)
            end_time = time.time()
            print(method,
                  " is done. Using {t:.4f}\n".format(t=end_time - start_time))

        scores = pd.concat(results, axis=1)
        z_scores = (scores - scores.mean()) / (scores.std() + 1e-4)
        scores[self.name] = z_scores.sum(axis=1)

        return scores

    def Select(self, scores):
        z_scores = scores[self.name].sort_values(ascending=False)

        return super().Select(z_scores)

    def what_matters(self):
        return self.scores_

    def Plotting(self):
        for method in self.kernels:
            self.kernels[method].Plotting()



from sklearn.decomposition import PCA
from sklearn.cross_decomposition import PLSSVD
import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator

# Todo: sklearn.pipeline.FeatureUnion


class feature_extension(BaseEstimator):

    def __init__(self,
                 alpha=0.9,
                 pca=True,
                 pls=False,
                 cross=True,
                 ratio=True,
                 name="extend "):

        self.alpha = alpha
        self.pca = pca
        self.pls = pls
        self.cross = cross
        self.ratio = ratio
        self.name = name

    def _set_up(self):
        if self.pca:
            self.pca = PCA()
        else:
            self.pca = None

        if self.pls:
            self.pls = PLSSVD(4)
        else:
            self.pls = None

    def fit(self, x, y):
        self._set_up()
        self.fitted_ = True
        results = [x.copy()]

        self.monotonic_feature = x.columns[np.logical_or(
            x.min() >= 0,
            x.max() <= 0)]
        self.monotonic_mean = x[self.monotonic_feature].mean()

        monotonic_x = x[self.monotonic_feature] / self.monotonic_mean

        if self.cross:
            x = monotonic_x.copy()
            column_names = monotonic_x.columns
            n = len(column_names)
            for i in range(n):
                for j in range(i + 1):
                    a = column_names[i]
                    b = column_names[j]
                    tmp = x[a] * x[b]
                    tmp.name = a + " * " + b

        if self.ratio:
            x = monotonic_x.copy()
            column_names = monotonic_x.columns
            n = len(column_names)
            for i in range(n):
                for j in range(i):
                    a = column_names[i]
                    b = column_names[j]
                    tmp = np.arctan2(x[a], x[b])
                    tmp.name = "arctan " + a + " / " + b

        x = results[0].copy()

        # PCA
        if self.pca:
            self.pca.fit(x)
            evr = self.pca.explained_variance_ratio_

            aev = 0
            counter = 0
            while aev < self.alpha and counter < len(evr):
                aev += evr[counter]
                counter += 1

            self.pca = PCA(counter)
            pcx = self.pca.fit_transform(x)
            pcx = pd.DataFrame(pcx,
                               index=x.index,
                               columns=[
                                   self.name + "pc" + str(i)
                                   for i in range(self.pca.n_components_)
                               ])

        if self.pls:
            if x.shape[1] < self.pls.n_components:
                self.pls.n_components = x.shape[1]
            self.pls.fit(x, y)

            plsx = self.pls.transform(x)
            plsx = pd.DataFrame(
                plsx,
                index=x.index,
                columns=[self.name + "pls" + str(i) for i in range(4)])

        return self

    def transform(self, x):
        results = [x.copy()]

        monotonic_x = x[self.monotonic_feature] / self.monotonic_mean

        if self.cross:
            x = monotonic_x.copy()
            column_names = monotonic_x.columns
            n = len(column_names)
            for i in range(n):
                for j in range(i + 1):
                    a = column_names[i]
                    b = column_names[j]
                    tmp = x[a] * x[b]
                    tmp.name = a + " * " + b
                    results.append(tmp)

        if self.ratio:
            x = monotonic_x.copy()
            column_names = monotonic_x.columns
            n = len(column_names)
            for i in range(n):
                for j in range(i):
                    a = column_names[i]
                    b = column_names[j]
                    tmp = np.arctan2(x[a], x[b])
                    tmp.name = "arctan " + a + " / " + b
                    results.append(tmp)

        x = results[0].copy()

        # PCA
        if self.pca:
            pcx = self.pca.transform(x)
            pcx = pd.DataFrame(pcx,
                               index=x.index,
                               columns=[
                                   self.name + "pc" + str(i)
                                   for i in range(self.pca.n_components_)
                               ])
            results.append(pcx)

        if self.pls:
            plsx = self.pls.transform(x)
            plsx = pd.DataFrame(
                plsx,
                index=x.index,
                columns=[self.name + "pls" + str(i) for i in range(4)])
            results.append(plsx)

        return pd.concat(results, axis=1)

    def fit_transform(self, x, y=None):
        self.fit(x, y)
        return self.transform(x)

    def report(self):
        state = {}
        return state



from . import Basic_tuner
from joblib import parallel_config

from sklearn.model_selection import StratifiedKFold
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.tree import DecisionTreeClassifier
from statsmodels.discrete.discrete_model import Logit
from sklearn.linear_model import LogisticRegression
from xgboost import XGBClassifier
from lightgbm import LGBMClassifier
from lightgbm import early_stopping as lgbm_early_stopping
from catboost import CatBoostClassifier, Pool

import shap

import numpy as np
from pandas import DataFrame


class Classification_tuner(Basic_tuner):
    """
    A subclass of Basic_tuner for classification task.

    """

    def __init__(self,
                 n_try,
                 n_cv,
                 target,
                 validate_penalty,
                 TT_coef,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def is_regression(self) -> bool:
        """
        No, this is not a regresion tuner or for regresion task.
        Returns:
            bool: False
        """

        return False

    def predict_proba(self, x):
        """
        The sklearn.base.BaseEstimator predict_prob api.

        Args:
            x (pandas.DataFrame): feature to extract information from.

        Returns:
            pd.DataFrame: prediction in prob, an array with shape (n_samples, n_classes)
        """
        return DataFrame(self.best_model.predict_proba(x),
                         index=x.index,
                         columns=self.y_mapping.classes_)


# linear model, elasticnet
class ElasticLogit_tuner(Classification_tuner):
    """
    Tuning a elasic net logistic regression.    
    [sklearn.linear_model.LogisticRegression](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html#sklearn.linear_model.LogisticRegression), reminds the choice of the algorithm depends on the penalty chosen and on (multinomial) multiclass support.    
    """

    def __init__(self,
                 n_try=25,
                 n_cv=5,
                 target="mcc",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed: int = None,
                 valid_seed: int = None,
                 optuna_seed: int = None):

        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

        # "saga" fast convergence is only guaranteed on features with approximately the same scale. You should do a feature-wise (between sample) normalization before fitting.
        self.solver = "saga"
        self.n_opt_jobs = 4

    def name(self):
        return "ElasticNetLogisticRegression"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html"
        return refer

    def parms_range(self) -> dict:
        return {
            "C": ('C', "float", 1e-6, 1e+2),
            "l1_ratio": ('l1_ratio', "float", -0.5, 1.5),
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "C": 1.0,
            "l1_ratio": 0.,
            "solver": "saga",
            "random_state": self.kernel_seed,
            "verbose": 0
        }
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                if par == "l1_ratio":
                    parms[par] = np.clip(
                        self.parms_range_sparser(trial, parms_to_tune[par]), 0,
                        1)
                else:
                    parms[par] = self.parms_range_sparser(
                        trial, parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
        lg = LogisticRegression(**parms)
        return lg

    def summary(self):
        """
        It is the way I found to cram a sklearn regression result into the statsmodel regresion.    
        The only reason to do this is that statsmodel provides R-style summary.    
        """
        if len(self.best_model.classes_) > 2:
            # Todo: multi-class classification
            raise TypeError(
                "multi-class classification summary not support yet. Please tell me why do you need that in a multi-class classification task"
            )
        else:
            # binary classification
            sm_logit = Logit(self.y, self.x).fit(
                disp=False,
                start_params=self.best_model.coef_.flatten(),
                maxiter=0,
                warn_convergence=False)
        print(sm_logit.summary())

    def _explainer(self, x):
        return shap.LinearExplainer(self.best_model, x)


# RF
class RandomForest_tuner(Classification_tuner):
    """
    Tuning a random forest model.    
    [sklearn.ensemble.RandomForestClassifier](https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html)
    """

    def __init__(self,
                 using_oob=True,
                 n_try=50,
                 n_cv=5,
                 target="mcc",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        """

        Args:
            using_oob (bool, optional): Using out of bag score as validation. Defaults to True.
        """
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

        self.using_oob = using_oob

    def name(self):
        return "RandomForest"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html"
        refer[
            self.name() +
            " publication"] = "https://www.stat.berkeley.edu/~breiman/randomforest2001.pdf"

        return refer

    def parms_range(self) -> dict:
        return {
            "n_estimators": ('n_estimators', "int", 32, 512),
            'min_samples_leaf':
            ('min_samples_leaf', "int", 1, round(np.sqrt(self.n_sample) / 2)),
            'ccp_alpha':
            ('ccp_alpha', "float", 1e-2 / self.n_sample, 1e+2 / self.n_sample),
            'max_samples': ('max_samples', "float", 0.4, 0.8),
            "max_depth":
            ("max_depth", "int", round(np.log2(self.n_sample) / 2),
             int(np.log2(self.n_sample)) + 2)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "bootstrap": True,
            "oob_score": self.using_oob,
            "n_jobs": -1,
            "random_state": self.kernel_seed,
            "verbose": 0,
        }

        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]

        rf = RandomForestClassifier(**parms)
        return rf

    def evaluate(self, trial, default=None, training=None):
        """
        RF has oob validation and we have it.
        
        """
        if default is None:
            default = self.default
        if training is None:
            training = self.training

        # create the model using from this trial
        classifier_obj = self.create_model(trial, default)

        if self.using_oob:
            # oob predict
            # there is a bug that default sklaern randomforest parallel_backend using thread where others use "loky", see joblib.parallel_backend.
            with parallel_config(backend='loky'):
                classifier_obj.fit(self.x,
                                   self.y,
                                   sample_weight=compute_sample_weight(
                                       class_weight="balanced", y=self.y))

            # oob prediction
            y_prob = classifier_obj.oob_decision_function_

            # tune a threshold via roc for Binary classification
            """
            if self.is_binary:
                if training:
                    fpr, tpr, thr = roc_curve(self.y, y_prob[:, 1])
                    self.thresholds[trial.number] = thr[abs(tpr -
                                                            fpr).argmax()]
            """

            # oob score
            ### manual scorer wraper.
            if self.metric_using_proba:
                if self.metric_name == "roc_auc":
                    # roc_auc can only be used on binary classification. Do not try ovr, ovo. forget them.
                    y_prob = y_prob[:, 1]
                score = self.metric._score_func(self.y, y_prob,
                                                **self.scorer_kargs)
            else:
                # revert to class symbols.
                """
                if self.is_binary:
                    if default:
                        t = 0.5
                    else:
                        t = self.thresholds[trial.number]
                    y_pred = classifier_obj.classes_[(y_prob[:, 1]
                                                      > t).astype(np.int16)]
                else:
                """
                y_pred = classifier_obj.classes_[y_prob.argmax(axis=-1)]
                score = self.metric._score_func(self.y, y_pred,
                                                **self.scorer_kargs)
            if not self.metric_great_better:
                # if not great is better, then multiply -1
                score *= -1
        else:
            with parallel_config(backend='loky'):
                score = super().evaluate(trial=trial,
                                         default=default,
                                         training=training)
        return score

    def _explainer(self, x):
        return shap.TreeExplainer(self.best_model)


# SVM
class SVM_tuner(Classification_tuner):
    """
    Tuning a support vector machine.    
    [sklearn.svm.SVC](https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html)
    """

    def __init__(self,
                 n_try=25,
                 n_cv=5,
                 target="mcc",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):

        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "SVM"

    def reference(self) -> dict[str, str]:
        """
        This function will return reference of this method in python dict.    
        If you want to access it in PineBioML api document, then click on the    >Expand source code     

        Returns:
            dict[str, str]: a dict of reference.
        """
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html"
        refer[
            self.name() +
            " publication"] = "https://citeseerx.ist.psu.edu/doc_view/pid/42e5ed832d4310ce4378c44d05570439df28a393"

        return refer

    def parms_range(self) -> dict:
        # scaling penalty: https://scikit-learn.org/stable/auto_examples/svm/plot_svm_scale_c.html#sphx-glr-auto-examples-svm-plot-svm-scale-c-py
        return {
            "kernel": ('kernel', "category", ["linear", "rbf",
                                              "sigmoid"], None),
            'C': ('C', "float", 1e-3 * np.sqrt(self.n_sample),
                  1e+2 * np.sqrt(self.n_sample))
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "kernel": "rbf",
            "random_state": self.kernel_seed,
            "probability": True,
            "gamma": "auto",
            'cache_size': 2000,
        }
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            #parms["random_state"] = self.kernel_seed_tape[trial.number]
        svm = SVC(**parms)
        return svm

    def _explainer(self, x):
        return shap.KernelExplainer(self.best_model.predict_proba, x)


# Todo: learning rate and number of iteration adjustment
# XGboost
class XGBoost_tuner(Classification_tuner):
    """
    Tuning a XGBoost classifier model.    
    [xgboost.XGBClassifier](https://xgboost.readthedocs.io/en/stable/python/python_api.html)

    ToDo:    
        1. sample imbalance. (we have temporary solution)    

    """

    def __init__(self,
                 n_try=75,
                 n_cv=5,
                 target="mcc",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        """

        Args:
            n_try (int, optional): The number of trials optuna should try. Defaults to 75.
            n_cv (int, optional): The number of folds to execute cross validation evaluation in iteration of optuna optimization. Defaults to 5.
            target (str, optional): The target of optuna optimization. Notice that is different from the training loss of model. Defaults to "mcc".
            kernel_seed (int, optional): Random seed for model. Defaults to None.
            valid_seed (int, optional): Random seed for cross validation. Defaults to None.
            optuna_seed (int, optional): Random seed for optuna's hyperparameter sampling. Defaults to None.
            validate_penalty (bool, optional): True to penalty the overfitting by difference between training score and cv score. Defaults to True.
        
        """
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "XGBoost"

    def reference(self) -> dict[str, str]:

        refer = super().reference()
        refer[self.name() +
              " document"] = "https://xgboost.readthedocs.io/en/stable/"
        refer[
            self.name() +
            " publication"] = "https://dl.acm.org/doi/10.1145/2939672.2939785"

        return refer

    def parms_range(self) -> dict:
        return {
            "n_estimators": ('n_estimators', "int", 4, 256),
            "max_depth":
            ("max_depth", "int", round(np.log2(self.n_sample) / 2),
             int(np.log2(self.n_sample)) * 2),
            "gamma": ('gamma', "float", 1e-4, 1e-2),
            "min_child_weight": ("min_child_weight", "float", 1,
                                 round(np.sqrt(self.n_sample) / 2)),
            "learning_rate": ('learning_rate', "float", 1e-2, 1.),
            "subsample": ('subsample', "float", 0.5, 1),
            "colsample_bytree": ('colsample_bytree', "float", 0.5, 1),
            "reg_lambda": ('reg_lambda', "float", 1e-4, 1),
            "reg_alpha": ('reg_alpha', "float", 1e-4, 1.)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "n_jobs": None,
            "random_state": self.kernel_seed,
            "verbosity": 0,
            "enable_categorical": True
        }
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]

            if training:
                # Adding early stopping callbacks
                parms["early_stopping_rounds"] = round(
                    parms["n_estimators"] * 0.1) + 2
            else:
                # not training, then overide the n_estimators by trial's early stopping point.
                parms["n_estimators"] = self.stop_points[trial.number]

        xgb = XGBClassifier(**parms)
        return xgb

    def using_earlystopping(self):
        return True

    def clr_best_iteration(self, classifier):
        return classifier.best_iteration

    def optimize_fit(self, clr, train_data, sample_weight, valid_data):

        train_x, train_y = train_data

        return clr.fit(train_x,
                       train_y,
                       sample_weight=sample_weight,
                       eval_set=[valid_data],
                       verbose=False)

    def _explainer(self, x):
        return shap.TreeExplainer(self.best_model)


# lightGBM
class LightGBM_tuner(Classification_tuner):
    """
    Tuning a LighGBM classifier model.    
    [lightgbm.LGBMClassifier](https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html)     

    ToDo:    
        Adapt to lightgbm train api. Each time lightgbm sklearn api was called, the data will be binned once and that is costy when number of features is large.
    """

    def __init__(self,
                 n_try=50,
                 n_cv=5,
                 target="mcc",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "LightGBM"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://lightgbm.readthedocs.io/en/latest/index.html"
        refer[
            self.name() +
            " publication"] = "https://proceedings.neurips.cc/paper_files/paper/2017/file/6449f44a102fde848669bdd9eb6b76fa-Paper.pdf"

        return refer

    def parms_range(self) -> dict:
        return {
            "n_estimators": ('n_estimators', "int", 4, 256),
            "max_depth":
            ("max_depth", "int", round(np.log2(self.n_sample) / 2),
             int(np.log2(self.n_sample)) * 2),
            "min_child_samples":
            ("min_child_samples", "int", 1, round(np.sqrt(self.n_sample) / 2)),
            "learning_rate": ('learning_rate', "float", 1e-2, 1.),
            "subsample": ('subsample', "float", 0.5, 1.),
            "colsample_bytree": ('colsample_bytree', "float", 0.5, 1.),
            "reg_lambda": ('reg_lambda', "float", 1e-4, 1),
            "reg_alpha": ('reg_alpha', "float", 1e-4, 1.)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "n_jobs": None,
            "random_state": self.kernel_seed,
            "verbosity": -1,
            "subsample_freq": 1
        }
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
            if not training:
                # not training, then overide the n_estimators by trial's early stopping point.
                parms["n_estimators"] = self.stop_points[trial.number]

        lgbm = LGBMClassifier(**parms)
        return lgbm

    def using_earlystopping(self):
        return True

    def clr_best_iteration(self, classifier):
        return classifier.best_iteration_

    def optimize_fit(self, clr, train_data, sample_weight, valid_data):
        train_x, train_y = train_data

        return clr.fit(train_x,
                       train_y,
                       sample_weight=sample_weight,
                       eval_set=[valid_data],
                       callbacks=[
                           lgbm_early_stopping(round(clr.n_estimators * 0.1) +
                                               2,
                                               verbose=False)
                       ])

    def _explainer(self, x):
        return shap.TreeExplainer(self.best_model)


# Adaboost
class AdaBoost_tuner(Classification_tuner):
    """
    Tuning a AdaBoost calssifier.    
    [sklearn.ensemble.AdaBoostClassifier](https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostClassifier.html)
    """

    def __init__(self,
                 n_try=25,
                 n_cv=5,
                 target="mcc",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)
        self.n_opt_jobs = 4

    def name(self):
        return "AdaBoost"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostClassifier.html"
        refer[
            self.name() +
            " publication: original version"] = "https://www.ee.columbia.edu/~sfchang/course/svia-F03/papers/freund95decisiontheoretic-adaboost.pdf"
        refer[
            self.name() +
            " publication: implemented version"] = "https://doi.org/10.4310/SII.2009.v2.n3.a8"

        return refer

    def parms_range(self) -> dict:
        return {
            "n_estimators": ('n_estimators', "int", 4, 64),
            "learning_rate": ('learning_rate', "float", 1e-2, 1.)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "random_state": self.kernel_seed,
        }
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
        ada = AdaBoostClassifier(**parms)
        return ada

    def _explainer(self, x):
        return shap.KernelExplainer(self.best_model.predict_proba, x)


# DT
class DecisionTree_tuner(Classification_tuner):
    """
    Tuning a DecisionTree calssifier.    
    [sklearn.tree.DecisionTreeClassifier](https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html)
    """

    def __init__(self,
                 n_try=25,
                 n_cv=5,
                 target="mcc",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)
        self.n_opt_jobs = 4

    def name(self):
        return "DecisionTree"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html"
        refer[
            self.name() +
            " publication"] = "https://www.taylorfrancis.com/books/mono/10.1201/9781315139470/classification-regression-trees-leo-breiman-jerome-friedman-olshen-charles-stone"

        return refer

    def parms_range(self) -> dict:
        return {
            "max_depth":
            ("max_depth", "int", round(np.log2(self.n_sample) / 2),
             int(np.log2(self.n_sample)) + 2),
            "min_samples_leaf":
            ('min_samples_leaf', "int", 1, round(np.sqrt(self.n_sample) / 2)),
            'ccp_alpha':
            ('ccp_alpha', "float", 1e-2 / self.n_sample, 1e+2 / self.n_sample),
        }

    def create_model(self, trial, default=False, training=False):
        parms = {"random_state": self.kernel_seed}
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
        DT = DecisionTreeClassifier(**parms)
        return DT

    def _explainer(self, x):
        return shap.TreeExplainer(self.best_model)


# catboost
class CatBoost_tuner(Classification_tuner):
    """
    Tuning a CatBoost classifier model.    
    [catboost.CatBoostClassifier](https://catboost.ai/en/docs/concepts/python-reference_catboostclassifier)     

    ToDo:    
        1. compare with optuna.integration.CatBoost...    
    """

    def __init__(self,
                 n_try=50,
                 n_cv=5,
                 target="mcc",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):

        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "CatBoost"

    def reference(self) -> dict[str, str]:

        refer = super().reference()
        refer[self.name() + " document"] = "https://catboost.ai/en/docs/"
        refer[
            self.name() +
            " publication"] = "https://proceedings.neurips.cc/paper_files/paper/2018/file/14491b756b3a51daac41c24863285549-Paper.pdf"

        return refer

    def parms_range(self) -> dict:
        return {
            "n_estimators": ('n_estimators', "int", 4, 256),
            "learning_rate": ('learning_rate', "float", 1e-1, 1.),
            "max_depth":
            ("max_depth", "int", round(np.log2(self.n_sample) / 2),
             int(np.log2(self.n_sample)) * 2),
            "reg_lambda": ('reg_lambda', "float", 1e-3, 1.),
            "colsample_bylevel": ('colsample_bytree', "float", 0.5, 1.),
            "subsample": ('subsample', "float", 0.5, 1.)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "random_state": self.kernel_seed,
            "verbose": False,
        }
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
        cat = CatBoostClassifier(**parms)
        return cat

    def optimize_fit(self, clr, train_data, sample_weight, valid_data):
        train_x, train_y = train_data
        valid_x, valid_y = valid_data

        cat_features = list(train_x.columns[train_x.dtypes == "category"])
        pool_train = Pool(train_x,
                          train_y,
                          weight=sample_weight,
                          cat_features=cat_features)
        pool_valid = Pool(valid_x, valid_y, cat_features=cat_features)

        return clr.fit(
            pool_train,
            eval_set=pool_valid,
            verbose=False,
            early_stopping_rounds=round(clr.get_param('n_estimators') * 0.1) +
            2)

    def using_earlystopping(self):
        return True

    def clr_best_iteration(self, classifier):
        return classifier.get_best_iteration()

    def _explainer(self, x):
        return shap.TreeExplainer(self.best_model)
